;
;	This file contains source code of XM module player
;	To play a XM module using this player do as below:
;		InitPlayer();
;		if (LoadSong(&szModuleName, 0, hwnd, 44100, 3, 4) == 0)
;		 {
;		   if (PlaySong(0) == 0)
;		    {
;		      // blah, blah, blah
;		      StopSong();
;		    };
;		   UnloadSong();
;		 }
;	Function LoadSong() loads the XM unpacking patterns and samples
;	  and storing them into allocated memory. Last three parameters
;	  of the routine are: frequency (mixing rate), flags (bit0 - mix
;	  stereo, bit1 - mix to 16bit, bit2 - 1 second mixing buffer instead
;	  of 0.5 second, bit3 - use MMX if available) and volume multiplier,
;	  bit4 - create high priority mixing thread.
;	Function PlaySong() installs the MixingThread() function as a new
;	  thread which responds to sound buffer notification positions
;	  and mixes samples when necessary. It calls Mix1(2) function which
;	  does actual mixing and updates sounds by calling TrackSong().
;	  When the first parameter is zero, the module loops infinite
;	  number of times, otherwise the parameter specifies an Event to be
;	  signaled when the module is halted at the end.
;	Function StopSong() removes the mixing thread and UnloadSong()
;	  releases all allocated memory and DirectX objects.
;	Function GetSongPosition() returns position within current module
;	  (high word is order number and low word is row number).
;	Function GetSyncEventState() returns data of the last
;	  synchronization command (Wxx), or -1 if no such command have
;	  been encountered since the last call to this function.
;	Function SetDirectSound() allows you to supply your own DirectSound
;	  object in the first argument (NULL causes creating a new object).
;	  The second argument (true/false) specifies whether to create a
;	  primary sound buffer and override playback parameters or not.
;	  Use this function only before LoadSong() (or after UnloadSong()),
;	  otherwise the player may release your DirectSound object.
;

ideal
p586n
pmmx
model flat, stdcall

; Uncomment this to exclude MMX mixing routines
;_NOMMX = 0

NOCOMMCTRL	= 0
NOCOMMDLG	= 0
NODDE		= 0
NOWINSPOOL	= 0
include "windows.inc"

NODIRECT3D	= 0
NODIRECTDRAW	= 0
NODIRECTINPUT	= 0
NODSOUND3D	= 0
include "directx.inc"

include "xm.inc"
include "xmplayer.inc"

udataseg

SongLength	db ?
RestartOrder	db ?
Speed		db ?
BPM		db ?
BPMTickDistance dd ?
NumChannels	dd ?
VolScale	dd ?
GlobalVol	db ?

Order		db ?				; Current song position
PatternPtr	dd ?				; Address of next row

DirectSound	dd ?				; IDirectSound object address
PrimaryBuffer	dd ?				; Primary sound buffer
SoundBuffer	dd ?				; Secondary sound buffer
BufferNotify	dd ?				; IDirectSoundNotify obj addr
hMixingThread	dd ?				; Handle of mixing thread

PatAddrTable	dd 256 dup (?)			; Order table - pattern addr-s
PatLengthTable	db 256 dup (?)			; Lengths of patterns

SpeedCountdown	db ?
RowCountdown	db ?
NumRows 	db ?
ChannelCountdown db ?
TicksLeft	dd ?				; Left unmixed samples to next
						; frame
ChannelTable	CHANNEL 32 dup (?)
SampleTable	dd ?
SampleBuffer	dd ?
InstrTable	dd ?
PatternMem	dd ?
MixingBuffer	dd ?
InstrTableSize	dd ?
SampleTableSize dd ?

label Notifies
Event1Pos	dd ?
Event1		dd ?
Event2Pos	dd ?
Event2		dd ?

Events		dq ?		; Doublebuffering events

struc	SYNCBUF
	BufferID	dd ?
	NumEvents	dd ?
	Events		dq 32 dup (?)
ends

SyncBuf1	SYNCBUF ?	; Sync Event buffers
SyncBuf2	SYNCBUF ?
SyncBuf3	SYNCBUF ?
CurrentSyncBuf	dd ?		; Current Sync Event buffer
SyncSection	CRITICAL_SECTION ? ; Critical section for sync events
LastSyncData	dd ?		; Data of last sync command
SoundBufPos	dd ?		; Passed to TrackSong (pos within sndbuf)
SoundBufBase	dd ?		; Sndbuf part ptr

Buf1Positions	dd 8+8*103 dup (?)
Buf2Positions	dd 8+8*103 dup (?)
NewBufPositions dd 8+8*103 dup (?)
LastBeforeBuf	dd ?

StopEvent	dd ?

; Linear frequency table: Freq = 8363*2^(Tune/1536)/16 ;Tune = Note<<7+Finetune
; In fact: Freq(C-4)=8363, Freq(C-5)=2*8363, etc.
FreqTable	dd (MAXNOTE+1)*128 dup (?)

; Table of vibratos
VibTable	dd 4*256 dup (?)

dataseg

IDText		db 'Extended Module: '
Flags		db flagNotLoaded
Flags2		db 0

Define_IID_IDirectSoundNotify

XlateEffect	db 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, -1, 12	;  0 -	C
		db 13, -1, 14, 15, -1, -1, -1, -1, -1, -1, -1	;  D - 17
		db -1, -1, -1, -1, -1, -1, -1, -1, 20, 21, -1	; 18 - 22
		db 22, 23, -1, -1, -1, -1, -1, -1		; E1 - E8
		db -1, 31, 32, -1, 34, -1, -1			; E9 - EF
XlateVolEffect	db 10, 10, 32, 31, -1, 4, 8, -1, -1, 3
VolEffectShift	db 0, 4, 0, 0, 4, 0, 4, 0, 4, 4

align
; Every function pointed by this array expects new dt|ef in ax and
; recent in dx and returns updated in ax	    H L
InitEffect	dd InitArpeggio, InitPorta, InitPorta, InitPorta
		dd InitVibrato, InitVolSlide, InitVolSlide, InitTremolo
		dd InitSetPan, InitSetSampOffs, InitVolSlide, 0
		dd InitSetVol, InitBreak, InitSetSpeed, InitSetGlobalVol
		dd 0, 0, 0, 0
		dd InitSync, InitExFinePorta, InitFinePortaUp, InitFinePortaDown
		dd 0, 0, 0, 0
		dd 0, 0, 0, InitFineVolSlUp
		dd InitFineVolSlDown, 0, InitNoteDelay

; Every fn ptd by the array expects dt in al
DoEffect	dd DoArpeggio, DoPortaUp, DoPortaDown, DoTonePorta
		dd DoVibrato, DoPortaVol, DoVibVol, DoTremolo
		dd -1, -1, DoVolSlide, -1
		dd -1, -1, -1, -1
		dd -1, -1, -1, -1
		dd -1, -1, -1, -1
		dd -1, -1, -1, -1
		dd -1, -1, -1, -1
		dd -1, -1, DoNoteDelay

; Mixer functions
Mixers		dd AMixM, AMixS,   MMixM, MMixS
		dd AIntM, AIntS,   MIntM, MIntS
ifndef _NOMMX
		dd 0,	  AMixMMX, 0,	  MMixMMX
		dd 0,	  AIntMMX, 0,	  MIntMMX
else
		dd 0,	  AMixS,   0,	  MMixS
		dd 0,	  AIntS,   0,	  MIntS
endif

_0M0M		dq 00007FFF00007FFFh
_MMMM		dq 7FFF7FFF7FFF7FFFh

OutputFreq	dd ?
IntStep 	dd ?
FracStep	dd ?
LeftVolume	dd ?
RightVolume	dd ?
ChannelAddr	dd ?
LoopLength	dd ?
EndPoint	dd ?
DoLoop		dd ?

; Supplied by SetDirectSound()
UserDirectSound dd 0
bCreatePrimary	dd 1

codeseg

; External function names for C
ifdef XMC
public _InitPlayer@0, _PlaySong@4, _StopSong@0, _GetSyncEventState@0
public _LoadSong@24, _UnloadSong@0, _GetSongPosition@0, _SetDirectSound@8
endif

;-----------------------------------------------------------------------------
; DoInstrumentEffects macro processes volume and panning envelopes, vibrato
; and fadeout
;-----------------------------------------------------------------------------

macro	DoInstrumentEffects
local _NoVolEnv, _NextVolPos, _NoPanEnv, _NextPanPos, _FadeoutNotExceeded
local _VolSust, _PanSust, _VibSweepLoop, _NoSweep, _VolEnvLoop, _PanEnvLoop
local _NoFadeout
	mov	edi, [(CHANNEL ebx).InstrAddr]	; Volume envelope
	test	[(INSTRUMENT edi).Flags], flagVolEnvelope
	jz	_NoVolEnv
	xor	eax, eax
	mov	ax, [(CHANNEL ebx).VolEnvPos]
	test	[(CHANNEL ebx).Flags], flagEndSample
	jz	_VolSust
	cmp	ax, [(INSTRUMENT edi).VolEnvBorder]
	jc	_VolEnvLoop
	cmp	[((INSTRUMENT edi).VolEnv)+eax-1], 0
	jnz	_NoVolEnv
	and	[(CHANNEL ebx).Flags], not flagSample
	jmp	_NoFadeout
_VolSust:
	cmp	ax, [(INSTRUMENT edi).VolSustain]
	jz	_NoVolEnv
_VolEnvLoop:
	inc	ax
	cmp	ax, [(INSTRUMENT edi).VolEndP1]
	jnz	_NextVolPos
	mov	ax, [(INSTRUMENT edi).VolBegin]
_NextVolPos:
	mov	[(CHANNEL ebx).VolEnvPos], ax
	cmp	ax, [(INSTRUMENT edi).VolEnvBorder]
	jnc	_NoVolEnv
	mov	al, [((INSTRUMENT edi).VolEnv)+eax]
	mov	[(CHANNEL ebx).VolEnvVol], al

_NoVolEnv:					; Panning envelope
	test	[(INSTRUMENT edi).Flags], flagPanEnvelope
	jz	_NoPanEnv
	mov	ax, [(CHANNEL ebx).PanEnvPos]
	test	[(CHANNEL ebx).Flags], flagEndSample
	jz	_PanSust
	cmp	ax, [(INSTRUMENT edi).PanEnvBorder]
	jnc	_NoPanEnv
	jmp	_PanEnvLoop
_PanSust:
	cmp	ax, [(INSTRUMENT edi).PanSustain]
	jz	_NoPanEnv
_PanEnvLoop:
	inc	ax
	cmp	ax, [(INSTRUMENT edi).PanEndP1]
	jnz	_NextPanPos
	mov	ax, [(INSTRUMENT edi).PanBegin]
_NextPanPos:
	mov	[(CHANNEL ebx).PanEnvPos], ax
	cmp	ax, [(INSTRUMENT edi).PanEnvBorder]
	jnc	_NoVolEnv
	movzx	eax, ax
	mov	al, [((INSTRUMENT edi).PanEnv)+eax]
	mov	[(CHANNEL ebx).PanEnvPan], al

_NoPanEnv:					; Vibrato
	movzx	eax, [(CHANNEL ebx).VibPos]
	add	al, [(INSTRUMENT edi).VibSpeed]
	mov	[(CHANNEL ebx).VibPos], al
	mov	ecx, [(INSTRUMENT edi).VibWaveAddr]
	mov	eax, [eax*4+ecx]
	movzx	ecx, [(INSTRUMENT edi).VibDepth]
	imul	ecx
	sar	eax, 6				; Vibrato sweep
	cmp	[(CHANNEL ebx).VibSweepCount], 0
	jz	_NoSweep
	dec	[(CHANNEL ebx).VibSweepCount]
	jz	_NoSweep
	movzx	ecx, [(CHANNEL ebx).VibSweepPos]
	add	ecx, 0ffh
	movzx	edx, [(CHANNEL ebx).FadeinVib]
_VibSweepLoop:
	inc	edx
	sub	cl, [(INSTRUMENT edi).VibSweep]
	sbb	ch, 0
	jnz	_VibSweepLoop
	mov	[(CHANNEL ebx).FadeinVib], dl
	mov	[(CHANNEL ebx).VibSweepPos], cl
	imul	edx
	sar	eax, 8
_NoSweep:
	mov	[(CHANNEL ebx).VibTune], eax
						; Volume fadeout
	test	[(CHANNEL ebx).Flags], flagEndSample
	jz	_NoFadeout
	mov	eax, [(CHANNEL ebx).FadeoutVol]
	sub	eax, [(INSTRUMENT edi).Fadeout]
	jnc	_FadeoutNotExceeded
	xor	eax, eax
	and	[(CHANNEL ebx).Flags], not flagSample
_FadeoutNotExceeded:
	mov	[(CHANNEL ebx).FadeoutVol], eax
_NoFadeout:
endm

;-----------------------------------------------------------------------------
; UpdateSndBuffer macro updates volume, panning and frequency of current sndbuf
;-----------------------------------------------------------------------------

macro	UpdateSndBuffer
local _VolDn, _VolReady, _DontNegatePan, _SetFreq, _NotTooSmall, _NoPanning
local _PanLeft
	mov	al, [GlobalVol] 		; Volume
	mul	[(CHANNEL ebx).Volume]
	movzx	eax, ax
	movzx	edx, [(CHANNEL ebx).VolEnvVol]
	mul	edx
	mul	[(CHANNEL ebx).FadeoutVol]
	shld	edx, eax, 5
	mov	dh, [(CHANNEL ebx).DeltaVol]
	or	dh, dh
	js	_VolDn
	add	dl, dh
	cmp	dl, 64
	jna	_VolReady
	mov	dl, 64
	jmp	_VolReady
_VolDn: add	dl, dh
	jc	_VolReady
	mov	dl, 0
_VolReady:
	mov	dh, dl

	test	[Flags], flagStereo		; Panning
	jz	_NoPanning
	movzx	eax, [(CHANNEL ebx).Panning]
	mov	ecx, eax
	test	al, 80h
	jz	_DontNegatePan
	neg	al
_DontNegatePan:
	mov	ah, [(CHANNEL ebx).PanEnvPan]
	sub	ah, 32
	imul	ah
	movsx	eax, ax
	sar	eax, 5
	add	eax, ecx
	sub	eax, 80h
	movzx	ecx, dl
	imul	eax, ecx
	sar	eax, 7
	js	_PanLeft
	sub	dh, al
	jmp	_NoPanning
_PanLeft:
	add	dl, al
_NoPanning:

	mov	cx, dx				; Final volume
	movzx	eax, cl
	mul	[VolScale]
	mov	[(CHANNEL ebx).RightVolume], eax
	movzx	eax, ch
	mul	[VolScale]
	mov	[(CHANNEL ebx).LeftVolume], eax

	mov	eax, [(CHANNEL ebx).Finetune]	; Frequency
	add	eax, [(CHANNEL ebx).DeltaTune]
	add	eax, [(CHANNEL ebx).VibTune]
	cmp	eax, 0
	jnl	_NotTooSmall
	xor	eax, eax
	jmp	_SetFreq
_NotTooSmall:
	cmp	eax, (MAXNOTE+1)*128
	jc	_SetFreq
	mov	eax, (MAXNOTE+1)*128-1
_SetFreq:
	mov	eax, [FreqTable+eax*4]
	xor	edx, edx
	div	[OutputFreq]
	mov	[(CHANNEL ebx).IntStep], eax
	xor	eax, eax
	div	[OutputFreq]
	mov	[(CHANNEL ebx).FracStep], eax
endm

;-----------------------------------------------------------------------------
;	TrackSong() callback itself
;-----------------------------------------------------------------------------

proc	TrackSong uses ebx esi edi
	mov	[LastSyncData], -1
	mov	al, [byte NumChannels]		; Update channel counter
	mov	[ChannelCountdown], al
	mov	ebx, O ChannelTable		; Channel structure pointer
	dec	[SpeedCountdown]		; Speed countdown
	jnz	_BetweenRows

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	Initialize channels for new row
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	mov	al, [Speed]			; Update speed counter
	mov	[SpeedCountdown], al

	dec	[RowCountdown]			; Was last row ?
	jnz	_GetNewRow

_NextOrder:
	inc	[Order] 			; Next order
	mov	al, [Order]
	cmp	al, [SongLength]		; Was last order ?
	jc	_NewPattern

	test	[Flags], flagLoopModule 	; Loop module ?
	jnz	_RestartModule

	or	[Flags2], flag2StopModule

_RestartModule:
	mov	al, [RestartOrder]		; Restart module
	mov	[Order], al

_NewPattern:
	movzx	eax, [Order]			; Get new pattern pointer
	mov	ecx, [PatAddrTable+eax*4]
	cmp	ecx, -1
	je	_NextOrder
	mov	[PatternPtr], ecx

	mov	al, [PatLengthTable+eax]	; Update row counter
	or	al, al
	jz	_NextOrder
	mov	[RowCountdown], al
	mov	[NumRows], al

_GetNewRow:
	and	[Flags], not flagBreak
	mov	esi, [PatternPtr]		; Pattern pointer

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	Next channel
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

_NewRowNextChannel:
	test	[byte esi], flagResetEffect2	; Delay note if needed
	jz	_NoNoteDelay
	cmp	[byte esi+6], 34
	jnz	_NoNoteDelay
	cmp	[byte esi+7], 0
	jz	_NoNoteDelay
	test	[byte esi], flagResetSample+flagResetNote+flagResetVolume
	jz	_NoNoteDelay
	test	[byte esi], flagSample
	jnz	_ProcessEffects
	test	[(CHANNEL ebx).Flags], flagSample
	jnz	_ProcessEffects
_NoNoteDelay:

	call	ProcessNoteSampleAndVolume	; Process new row data
_ProcessEffects:
	call	ProcessEffects

_EndNewRow:
	test	[(CHANNEL ebx).Flags], flagSample
	jz	_NoUpdate
	UpdateSndBuffer
_NoUpdate:
	add	ebx, size CHANNEL		; Next channel
	add	esi, 8
	dec	[ChannelCountdown]
	jnz	_NewRowNextChannel
						; OK. All channels done
	test	[Flags], flagBreak
	jnz	_SkipPatPtrUpdate
	mov	[PatternPtr], esi
_SkipPatPtrUpdate:

	cmp	[LastSyncData], -1		; Add sync event
	je	_NoSyncEventOccured
	mov	eax, [SyncBuf3.NumEvents]
	cmp	eax, 32
	jnb	_NoSyncEventOccured
	mov	ecx, [SoundBufPos]
	mov	[dword (SyncBuf3.Events)+eax*8], ecx
	mov	ecx, [LastSyncData]
	mov	[dword (SyncBuf3.Events)+eax*8+4], ecx
	inc	eax
	mov	[SyncBuf3.NumEvents], eax
_NoSyncEventOccured:

	xor	eax, eax
	ret

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	Process samples and their effects between rows
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

_BetweenRows:
_BetwRowsNextChannel:				; Check if there is a sample
	test	[(CHANNEL ebx).Flags], flagSample
	jz	_NoInstrEff
	DoInstrumentEffects

_NoInstrEff:					; Is effect 1 ?
	test	[(CHANNEL ebx).Flags], flagEff1On
	jz	_TryEff2
	movzx	ecx, [(CHANNEL ebx).Effect1]	; Do it
	xor	eax, eax
	xchg	al, ch
	call	[DoEffect+ecx*4]

_TryEff2:					; Is effect 2 ?
	test	[(CHANNEL ebx).Flags], flagEff2On
	jz	_SndBuffer
	movzx	ecx, [(CHANNEL ebx).Effect2]	; Do it
	xor	eax, eax
	xchg	al, ch
	call	[DoEffect+ecx*4]

_SndBuffer:					; Finish channel
	test	[(CHANNEL ebx).Flags], flagSample
	jz	_NextChannel
	UpdateSndBuffer
_NextChannel:
	add	ebx, size CHANNEL
	dec	[ChannelCountdown]
	jnz	_BetwRowsNextChannel

	xor	eax, eax
	ret
endp

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	Process note, sample and volume
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc	ProcessNoteSampleAndVolume
	test	[byte esi], flagResetNote	; New pat note ?
	jz	_ResetSample
	mov	[(CHANNEL ebx).VibEPos], 0	; Retrig vibrato and tremolo
	mov	[(CHANNEL ebx).TremEPos], 0
	and	[(CHANNEL ebx).Flags], not flagEndNote
	mov	al, [esi+1]			; Get new pat note
	mov	[(CHANNEL ebx).Note], al
	cmp	al, 96
	jnz	_ResetSample
	or	[(CHANNEL ebx).Flags], flagEndNote
						; Stop sample / env loops
	test	[(CHANNEL ebx).Flags], flagSample
	jz	_ResetSample
	mov	eax, [(CHANNEL ebx).InstrAddr]
	test	[(INSTRUMENT eax).Flags], flagVolEnvelope
	jz	_StopPlayingSample
	or	[(CHANNEL ebx).Flags], flagEndSample
	and	[(CHANNEL ebx).Flags], not (flagEff1On+flagEff2On)
	jmp	_ResetSample
_StopPlayingSample:
	mov	[(CHANNEL ebx).Flags], flagEndNote

_ResetSample:
	test	[byte esi], flagResetSample	; New sample ?
	jnz	_GetNewSample
	test	[(CHANNEL ebx).Flags], flagSample
	jz	_ResetVolume
	DoInstrumentEffects
	jmp	_ResetVolume
_GetNewSample:
	movzx	eax, [byte esi+2]
	cmp	al, -1				; Perform action on bare reset
	jnz	_JustResetSample
_ChkEndNote:
	test	[(CHANNEL ebx).Flags], flagEndNote
	jnz	_ResetVolume
	movzx	eax, [(CHANNEL ebx).Instr]
_JustResetSample:
	and	[(CHANNEL ebx).Flags], not flagSample
	mov	[(CHANNEL ebx).Instr], al	; Update current instrument num
	imul	eax, size INSTRUMENT		; Get sample for current note
	cmp	eax, [InstrTableSize]
	jnc	_ResetVolume
	add	eax, [InstrTable]
	mov	[(CHANNEL ebx).InstrAddr], eax
	movzx	ecx, [(CHANNEL ebx).Note]
	movzx	eax, [((INSTRUMENT eax).Samples)+ecx]
	imul	eax, size SAMPLE
	cmp	eax, [SampleTableSize]
	jnc	_ResetVolume
	add	eax, [SampleTable]
	mov	[(CHANNEL ebx).SampleAddr], eax

	mov	ecx, [(SAMPLE eax).SampleBuf]	; Play new sample
	or	ecx, ecx
	jz	_ResetVolume
	mov	[(CHANNEL ebx).SamplePos], ecx
	mov	[(CHANNEL ebx).FracPos], 0
	or	[(CHANNEL ebx).Flags], flagSample
	and	[(CHANNEL ebx).Flags], not flagEndSample
						; Reset instrument effects
	mov	cl, [(SAMPLE eax).Panning]
	mov	[(CHANNEL ebx).Panning], cl
	mov	ecx, [(CHANNEL ebx).InstrAddr]
	mov	[(CHANNEL ebx).VolEnvPos], 0
	mov	[(CHANNEL ebx).PanEnvPos], 0
	mov	[(CHANNEL ebx).VibPos], 0
	mov	dl, [(INSTRUMENT ecx).VibSweep]
	mov	[(CHANNEL ebx).VibSweepCount], dl
	mov	[(CHANNEL ebx).VibSweepPos], 0
	mov	[(CHANNEL ebx).FadeinVib], 0
	mov	[(CHANNEL ebx).VibTune], 0
	mov	dl, 40h
	test	[(INSTRUMENT ecx).Flags], flagVolEnvelope
	jz	_No1stVolEnv
	mov	dl, [(INSTRUMENT ecx).VolEnv]
_No1stVolEnv:
	mov	[(CHANNEL ebx).VolEnvVol], dl
	mov	dl, 20h
	test	[(INSTRUMENT ecx).Flags], flagPanEnvelope
	jz	_No1stPanEnv
	mov	dl, [(INSTRUMENT ecx).PanEnv]
_No1stPanEnv:
	mov	[(CHANNEL ebx).PanEnvPan], dl
	mov	[(CHANNEL ebx).FadeoutVol], 8000h
	movzx	edx, [(CHANNEL ebx).Note]
	shl	edx, 7
	add	edx, [(SAMPLE eax).Finetune]
	mov	[(CHANNEL ebx).Finetune], edx

_ResetVolume:
	test	[byte esi], flagResetVolume	; New volume ?
	jz	_NoteSampleAndVolumeProcessed
	mov	al, [esi+3]			; Get new volume
	cmp	al, -1
	jnz	_SetVolume
	inc	al				  ;
	test	[(CHANNEL ebx).Flags], flagSample ; Safe
	jz	_SetVolume			  ;
	mov	edx, [(CHANNEL ebx).SampleAddr]
	mov	al, [(SAMPLE edx).Volume]
_SetVolume:
	mov	[(CHANNEL ebx).Volume], al
_NoteSampleAndVolumeProcessed:
	ret
endp

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	Process effects
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc	ProcessEffects
	mov	[(CHANNEL ebx).DeltaTune], 0	; These are affected by effects
	mov	[(CHANNEL ebx).DeltaVol], 0

	and	[(CHANNEL ebx).Flags], not flagEff1On
	test	[byte esi], flagResetEffect1	; New effect 1 ?
	jz	_NewEffect2
	mov	ax, [esi+4]			; Reset effect 1
	mov	dx, [(CHANNEL ebx).Effect1]
	movzx	ecx, al
	cmp	ecx, 35 			;
	jnb	_NewEffect2			; Anti-crash
	cmp	[InitEffect+ecx*4], 0		;
	je	_NewEffect2			;
	cmp	[DoEffect+ecx*4], -1
	jz	_Eff1InitOnly
	or	[(CHANNEL ebx).Flags], flagEff1On
_Eff1InitOnly:
	call	[InitEffect+ecx*4]
	mov	[(CHANNEL ebx).Effect1], ax

_NewEffect2:
	and	[(CHANNEL ebx).Flags], not flagEff2On
	test	[byte esi], flagResetEffect2	; New effect 2 ?
	jz	_EffectsProcessed
	mov	ax, [esi+6]			; Reset effect 2
	mov	dx, [(CHANNEL ebx).Effect2]
	movzx	ecx, al
	cmp	ecx, 35 			;
	jnb	_EffectsProcessed		; Anti-crash
	cmp	[InitEffect+ecx*4], 0		;
	je	_EffectsProcessed		;
	cmp	[DoEffect+ecx*4], -1
	jz	_Eff2InitOnly
	or	[(CHANNEL ebx).Flags], flagEff2On
_Eff2InitOnly:
	call	[InitEffect+ecx*4]
	mov	[(CHANNEL ebx).Effect2], ax
_EffectsProcessed:
	ret
endp

;-----------------------------------------------------------------------------
;	Effects !
;-----------------------------------------------------------------------------

proc	InitArpeggio
	mov	[(CHANNEL ebx).ArpPos], 3
	ret
endp

proc	DoArpeggio
	xor	ecx, ecx
	mov	dl, [(CHANNEL ebx).ArpPos]
	dec	dl
	jnz	_NoArpWrap
	mov	dl, 3
	jmp	_ArpEnd
_NoArpWrap:
	cmp	dl, 2
	jnz	_NoArpShift
	shr	al, 4
_NoArpShift:
	and	al, 0Fh
	mov	ecx, eax
	shl	ecx, 7
_ArpEnd:
	mov	[(CHANNEL ebx).ArpPos], dl
	mov	[(CHANNEL ebx).DeltaTune], ecx
	ret
endp

proc	InitPorta
	movzx	ecx, al
	dec	ecx
	or	ah, ah
	jnz	_NoZeroPorta
	mov	ah, [((CHANNEL ebx).Porta)+ecx]
_NoZeroPorta:
	mov	[((CHANNEL ebx).Porta)+ecx], ah
	ret
endp

proc	DoPortaUp
	shl	eax, 3
	add	[(CHANNEL ebx).Finetune], eax
	ret
endp

proc	DoPortaDown
	shl	eax, 3
	sub	[(CHANNEL ebx).Finetune], eax
	ret
endp

proc	DoTonePorta
	test	[(CHANNEL ebx).Flags], flagSample
	jz	_NoTonePorta
	shl	eax, 3
	movzx	ecx, [(CHANNEL ebx).Note]
	shl	ecx, 7
	mov	edx, [(CHANNEL ebx).SampleAddr]
	add	ecx, [(SAMPLE edx).Finetune]
	mov	edx, [(CHANNEL ebx).Finetune]
	cmp	ecx, edx
	jz	_NoTonePorta
	jc	_TonePortaDown
	add	edx, eax
	cmp	edx, ecx
	jng	_TonePortaEnd
	mov	edx, ecx
	jmp	_TonePortaEnd
_TonePortaDown:
	sub	edx, eax
	cmp	edx, ecx
	jnl	_TonePortaEnd
	mov	edx, ecx
_TonePortaEnd:
	mov	[(CHANNEL ebx).Finetune], edx
_NoTonePorta:
	ret
endp

proc	InitVibrato
	test	ah, 0Fh
	jnz	_NoVibDepthChange
	mov	cl, [(CHANNEL ebx).Vib]
	and	cl, 0Fh
	and	ah, 0F0h
	or	ah, cl
_NoVibDepthChange:
	test	ah, 0F0h
	jnz	_NoVibRateChange
	mov	cl, [(CHANNEL ebx).Vib]
	and	cl, 0F0h
	and	ah, 0Fh
	or	ah, cl
_NoVibRateChange:
	mov	[(CHANNEL ebx).Vib], ah
	ret
endp

proc	DoVibrato
	movzx	ecx, [(CHANNEL ebx).VibEPos]
	mov	dl, al
	shr	dl, 4
	shl	dl, 2
	add	cl, dl
	mov	[(CHANNEL ebx).VibEPos], cl
	movzx	edx, [(CHANNEL ebx).VibEWave]
	shl	edx, 10
	and	al, 0Fh
	imul	[VibTable+edx+ecx*4]
	sar	eax, 4
	neg	eax
	mov	[(CHANNEL ebx).DeltaTune], eax
	ret
endp

proc	DoPortaVol
	push	eax
	movzx	eax, [((CHANNEL ebx).Porta)+2]
	call	DoTonePorta
	pop	eax
	jmp	DoVolSlide
endp

proc	DoVibVol
	push	eax
	movzx	eax, [(CHANNEL ebx).Vib]
	call	DoVibrato
	pop	eax
	jmp	DoVolSlide
endp

proc	InitTremolo
	test	ah, 0Fh
	jnz	_NoTremDepthChange
	mov	cl, [(CHANNEL ebx).Tremolo]
	and	cl, 0Fh
	and	ah, 0F0h
	or	ah, cl
_NoTremDepthChange:
	test	ah, 0F0h
	jnz	_NoTremRateChange
	mov	cl, [(CHANNEL ebx).Tremolo]
	and	cl, 0F0h
	and	ah, 0Fh
	or	ah, cl
_NoTremRateChange:
	mov	[(CHANNEL ebx).Tremolo], ah
	ret
endp

proc	DoTremolo
	movzx	ecx, [(CHANNEL ebx).TremEPos]
	mov	dl, al
	shr	dl, 4
	shl	dl, 2
	add	cl, dl
	mov	[(CHANNEL ebx).TremEPos], cl
	movzx	edx, [(CHANNEL ebx).TremEWave]
	shl	edx, 10
	and	al, 0Fh
	imul	[VibTable+edx+ecx*4]
	sar	eax, 6
	neg	al
	mov	[(CHANNEL ebx).DeltaVol], al
	ret
endp

proc	InitSetPan
	mov	[(CHANNEL ebx).Panning], ah
	ret
endp

proc	InitSetSampOffs
	or	ah, ah
	jnz	_UpdateSampOffs
	mov	ah, [(CHANNEL ebx).SampleOffs]
_UpdateSampOffs:
	mov	[(CHANNEL ebx).SampleOffs], ah
	test	[(CHANNEL ebx).Flags], flagSample
	jz	_NoSampOffs
	lea	ecx, [eax+eax]
	and	ecx, 0001FE00h
	mov	edx, [(CHANNEL ebx).SampleAddr]
	add	ecx, [(SAMPLE edx).SampleBuf]
	cmp	ecx, [(SAMPLE edx).LoopEnd]
	jc	_SetSampOffs
	and	[(CHANNEL ebx).Flags], not flagSample
	jmp	_NoSampOffs
_SetSampOffs:
	mov	[(CHANNEL ebx).SamplePos], ecx
_NoSampOffs:
	ret
endp

proc	InitVolSlide
	or	ah, ah
	jnz	_NoVolSlChange
	mov	ah, [(CHANNEL ebx).VolSlide]
_NoVolSlChange:
	mov	[(CHANNEL ebx).VolSlide], ah
	ret
endp

proc	DoVolSlide
	mov	ah, [(CHANNEL ebx).Volume]
	test	al, 0F0h
	jz	_VolSlDown
	test	al, 0Fh
	jnz	_NoVolSlide
	shr	al, 4
	add	ah, al
	cmp	ah, 40h
	jna	_SaveVolSlide
	mov	ah, 40h
	jmp	_SaveVolSlide
_VolSlDown:
	sub	ah, al
	jnc	_SaveVolSlide
	mov	ah, 0
_SaveVolSlide:
	mov	[(CHANNEL ebx).Volume], ah
_NoVolSlide:
	ret
endp

proc	InitSetVol
	cmp	ah, 40h
	jna	_VolOkToSet
	mov	ah, 40h
_VolOkToSet:
	mov	[(CHANNEL ebx).Volume], ah
	ret
endp

proc	InitBreak
	test	[Flags], flagBreak
	jnz	_Break_GetPattern
	or	[Flags], flagBreak
	inc	[Order]
	movzx	ecx, [Order]
	cmp	cl, [SongLength]
	jc	_Break_GetPattern
	test	[Flags], flagLoopModule
	jnz	_Break_Restart
	dec	[Order]
	dec	ecx
	mov	[RowCountdown], 1
	jmp	_Break_Done
_Break_Restart:
	movzx	ecx, [RestartOrder]
	mov	[Order], cl
_Break_GetPattern:
	movzx	ecx, [Order]
	mov	dl, [PatLengthTable+ecx]
	cmp	ah, dl
	jc	_Break_RowOk
	mov	ah, 0
_Break_RowOk:
	inc	dl
	sub	dl, ah
	mov	[RowCountdown], dl
	mov	edx, [PatAddrTable+ecx*4]
	movzx	ecx, ah
	lea	edx, [edx+ecx*8]
	mov	[PatternPtr], edx
_Break_Done:
	ret
endp

proc	InitSetSpeed
	cmp	ah, 20h
	jnc	_SetBPM
	mov	[Speed], ah
	mov	[SpeedCountdown], ah
_SetBPM:
	ret
endp

proc	InitSetGlobalVol
	cmp	ah, 40h
	jna	_GVolOkToSet
	mov	ah, 40h
_GVolOkToSet:
	mov	[GlobalVol], ah
	ret
endp

proc	InitSync
	movzx	ecx, ah
	mov	[LastSyncData], ecx
	ret
endp

proc	InitExFinePorta
	xor	ecx, ecx
	mov	cl, ah
	shr	ah, 4
	jz	_exfpNoExPorta
	cmp	ah, 2
	jz	_exfpDown
	jnc	_exfpNoExPorta

_exfpUp:
	and	cl, 0Fh
	jnz	_exfpNonZeroUp
	mov	cl, [(CHANNEL ebx).ExFinePortaUp]
_exfpNonZeroUp:
	mov	[(CHANNEL ebx).ExFinePortaUp], cl
	add	ecx, ecx
	add	[(CHANNEL ebx).Finetune], ecx
	ret

_exfpDown:
	and	cl, 0Fh
	jnz	_exfpNonZeroDown
	mov	cl, [(CHANNEL ebx).ExFinePortaDown]
_exfpNonZeroDown:
	mov	[(CHANNEL ebx).ExFinePortaDown], cl
	add	ecx, ecx
	sub	[(CHANNEL ebx).Finetune], ecx
_exfpNoExPorta:
	ret
endp

proc	InitFinePortaUp
	or	ah, ah
	jnz	_NoZeroFinePortaUp
	mov	ah, [(CHANNEL ebx).FinePortaUp]
_NoZeroFinePortaUp:
	mov	[(CHANNEL ebx).FinePortaUp], ah
	movzx	ecx, ah
	shl	ecx, 3
	add	[(CHANNEL ebx).Finetune], ecx
	ret
endp

proc	InitFinePortaDown
	or	ah, ah
	jnz	_NoZeroFinePortaDown
	mov	ah, [(CHANNEL ebx).FinePortaDown]
_NoZeroFinePortaDown:
	mov	[(CHANNEL ebx).FinePortaDown], ah
	movzx	ecx, ah
	shl	ecx, 3
	sub	[(CHANNEL ebx).Finetune], ecx
	ret
endp

proc	InitFineVolSlUp
	mov	cl, [(CHANNEL ebx).Volume]
	add	cl, ah
	cmp	cl, 40h
	jna	_NoFVSUExceed
	mov	cl, 40h
_NoFVSUExceed:
	mov	[(CHANNEL ebx).Volume], cl
	ret
endp

proc	InitFineVolSlDown
	mov	cl, [(CHANNEL ebx).Volume]
	sub	cl, ah
	jnc	_NoFVSDExceed
	mov	cl, 00h
_NoFVSDExceed:
	mov	[(CHANNEL ebx).Volume], cl
	ret
endp

proc	InitNoteDelay
	mov	[(CHANNEL ebx).ndCount], ah
	mov	[(CHANNEL ebx).ndPattern], esi
	ret
endp

proc	DoNoteDelay
	dec	[(CHANNEL ebx).ndCount]
	jnz	_NoDelayedNote
	mov	esi, [(CHANNEL ebx).ndPattern]
	call	ProcessNoteSampleAndVolume
	call	ProcessEffects
	and	[(CHANNEL ebx).Flags], not flagEff2On
_NoDelayedNote:
	ret
endp

;-----------------------------------------------------------------------------
;	MixingThread() performs mixing into the sound buffer when necessary
;-----------------------------------------------------------------------------

proc	MixingThread
_WaitEvent:
	call	_WaitForMultipleObjects, 2, O Events, 0, INFINITE
	test	[Flags2], flag2Stopping
	jnz	_StopModule
	test	[Flags2], flag2StopModule
	jz	_DontStop
	or	[Flags2], flag2Stopping
_DontStop:
	cmp	eax, WAIT_OBJECT_0 + 1	; WAIT_OBJECT_0 == 0
	ja	_WaitEvent
	jz	_Mix1
	call	Mix2
	jmp	_WaitEvent
_Mix1:	call	Mix1
	jmp	_WaitEvent
_StopModule:
	call	_EnterCriticalSection, O SyncSection
	mov	eax, [SoundBuffer]		; Stop module
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Stop], eax
	mov	eax, [SoundBuffer]
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).SetCurrentPosition], eax, 0
	call	_SetEvent, [StopEvent]
	and	[Flags2], not (flag2StopModule+flag2Stopping)
	call	_LeaveCriticalSection, O SyncSection
	jmp	_WaitEvent
endp

align
proc	Mix1
	xor	ebx, ebx
	jmp	_MixN
endp

align
proc	Mix2
	mov	ebx, [Event2Pos]

_MixN:	mov	eax, [CurrentSyncBuf]		; Prepare sync events
	cmp	eax, O SyncBuf3
	jb	__NoMoveSyncEvents
	call	_EnterCriticalSection, O SyncSection
	mov	eax, [SyncBuf2.NumEvents]
	or	eax, eax
	jz	__NoMoveBuf2
	mov	[SyncBuf1.NumEvents], 1
	mov	eax, [dword (SyncBuf2.Events)+eax*8+4 -8]
	mov	[dword (SyncBuf1.Events)+4], eax
__NoMoveBuf2:
	push	esi
	push	edi
	mov	esi, O SyncBuf3
	mov	edi, O SyncBuf2
	mov	ecx, (size SYNCBUF) shr 2
    rep movsd
	pop	edi
	pop	esi
	mov	[CurrentSyncBuf], O SyncBuf2
	call	_LeaveCriticalSection, O SyncSection
__NoMoveSyncEvents:
	xor	eax, eax
	or	ebx, ebx
	setnz	al
	inc	eax
	mov	[SyncBuf3.BufferID], eax
	mov	[SyncBuf3.NumEvents], 0
	mov	[SoundBufBase], ebx
	mov	[NewBufPositions], 0
	mov	eax, [LastBeforeBuf]
	mov	[NewBufPositions+4], eax

	push	eax				; Get current status
	mov	eax, esp
	mov	esi, [SoundBuffer]
	mov	edi, [esi]
	call	[(IDirectSoundBuffer edi).GetStatus], esi, eax
	pop	eax
	test	eax, DSBSTATUS_BUFFERLOST
	jz	_DontRestore
						; Try to restore lost buffer
	call	[(IDirectSoundBuffer edi).Restore], esi
	or	eax, eax
	jz	_DontRestore
	ret

_DontRestore:
	sub	esp, 16 			; Lock the sound buffer
	lea	eax, [esp+12]
	push	0	; <- param 7
	push	eax	; <- param 6
	sub	eax, 4
	push	eax	; <- param 5
	sub	eax, 4
	push	eax	; <- param 4
	sub	eax, 4
	push	eax	; <- param 3
	call	[(IDirectSoundBuffer edi)._Lock], esi, ebx, [Event2Pos]
	or	eax, eax
	jz	_BufferLocked
	add	esp, 12
	ret
_BufferLocked:

	mov	edi, [esp]			; Get sound buffer addr and size
	mov	ebx, [esp+4]
	test	[Flags], flagStereo
	jz	_MixNextPart16
	shr	ebx, 1
_MixNextPart16:
	test	[Flags], flagMix16
	jz	_MixNextPart
	shr	ebx, 1

_MixNextPart:
	test	[Flags2], flag2StopModule	; Zero buffer if end of module
	jnz	_ZeroBuffer

	mov	ecx, ebx			; Get next unmixed part size
	cmp	ecx, [TicksLeft]
	jna	_SmallPartLeft
	mov	ecx, [TicksLeft]
_SmallPartLeft:
	sub	[TicksLeft], ecx
	sub	ebx, ecx
	push	ebx

	mov	ebx, O ChannelTable		; Mix first channel
	mov	esi, [NumChannels]
	xor	eax, eax
_MixNextChannel:				; Is there a sample ?
	test	[(CHANNEL ebx).Flags], flagSample
	jz	_TryToMixNextChannel

	pusha					; Choose mixer
	test	[Flags], flagStereo
	setnz	al
	mov	edx, [(CHANNEL ebx).SampleAddr]
	movzx	esi, al 			; Prepare loop data
	mov	[DoLoop], 0
	test	[(SAMPLE edx).Flags], flagLoop
	jz	_LoopNotExist
	mov	eax, [(SAMPLE edx).LoopLength]
	shr	eax, 1
	mov	[LoopLength], eax
	mov	[DoLoop], 1
_LoopNotExist:

	mov	eax, [(CHANNEL ebx).LeftVolume]
	mov	[LeftVolume], eax
	mov	eax, [(CHANNEL ebx).RightVolume]
	mov	[RightVolume], eax

	test	[Flags], flagUseMMX		; MMX
	jz	_MMXReady
	cmp	ecx, 2
	jb	_MMXReady
	or	esi, 08h
	movq	mm3, [qword LeftVolume]
	packssdw mm3, mm3

_MMXReady:
	mov	eax, [(CHANNEL ebx).IntStep]
	mov	[IntStep], eax
	or	eax, eax
	jnz	_NoInterp
	or	esi, 04h
_NoInterp:
	mov	eax, [(CHANNEL ebx).FracStep]
	mov	[FracStep], eax

	mov	eax, esi			; Update proper registers
	mov	esi, [(CHANNEL ebx).SamplePos]
	shr	esi, 1
	mov	[ChannelAddr], ebx
	mov	edi, [MixingBuffer]
	mov	ebx, [(CHANNEL ebx).FracPos]
	mov	edx, [(SAMPLE edx).LoopEnd]
	shr	edx, 1
	mov	[EndPoint], edx

	cmp	[byte esp+28], 0		; Call mixer
	jnz	_AddMixer
	inc	[byte esp+28]
	or	al, 02h
_AddMixer:
	call	[Mixers+eax*4]
	popa
_TryToMixNextChannel:
	add	ebx, size CHANNEL
	dec	esi
	jnz	_MixNextChannel

	mov	esi, [MixingBuffer]		; Finish buffer
	test	[Flags], flagStereo
	jz	_NoFBCountShift
	shl	ecx, 1
_NoFBCountShift:
	test	[Flags], flagUseMMX
	jnz	_FinishBufMMX
	test	[Flags], flagMix16
	jz	_FinishBuf8

_FinishBuf16:					; 16bit shift and saturation
	lodsd
	sar	eax, 6 + 4	; 6 - volume, 4 - volscale
	cmp	eax, 00007FFFh
	jg	_16SatDn
	cmp	eax, 0FFFF8000h
	jl	_16SatUp
_16SatEnd:
	stosw
	loop	_FinishBuf16
	jmp	_PartFinished
_16SatDn:
	mov	eax, 00007FFFh
	jmp	_16SatEnd
_16SatUp:
	mov	eax, 0FFFF8000h
	jmp	_16SatEnd

_8SatDn:
	mov	eax, 7Fh			; 8bit shift and saturation
	jmp	_8SatEnd
_8SatUp:
	mov	eax, 80h
	jmp	_8SatEnd
_FinishBuf8:
	lodsd
	sar	eax, 6 + 4 + 8	; 6 - volume, 4 - volscale, 8 - down to 8bit
	cmp	eax, 007Fh
	jg	_8SatDn
	cmp	eax, 0FFFFFF80h
	jl	_8SatUp
_8SatEnd:
	xor	eax, 80h
	stosb
	loop	_FinishBuf8
	jmp	_PartFinished

_FinishBufMMX:					; 16bit shift & sat. with MMX
	mov	al, cl
	and	al, 07h
	mov	[byte DoLoop], al
	shr	ecx, 3
	jz	_MMXRemain

	; 1.125 cycles per sample
	movq	mm0, [esi]
_next:	movq	mm1, [esi+8]		; 1
	psrad	mm0, 6 + 4		; -
	movq	mm2, [esi+16]		; 2
	psrad	mm1, 6 + 4		; -
	movq	mm3, [esi+24]		; 3
	packssdw mm0, mm1		; -
	psrad	mm2, 6 + 4		; 4
	add	esi, 32 		; -
	movq	[edi], mm0		; 5
	psrad	mm3, 6 + 4		; -
	packssdw mm2, mm3		; 6
	add	edi, 16 		; -
	movq	mm0, [esi]		; 7
	movq	[edi-8], mm2		; 8
	dec	ecx			; 9
	jnz	_next			; -

_MMXRemain:
	mov	ecx, [DoLoop]
	or	ecx, ecx
	jnz	_FinishBuf16
_PartFinished:

	cmp	[TicksLeft], 0			; Update channels and mix next
	jnz	_NoChannelUpdate		; part
	push	edi
	mov	eax, edi			; Position within sndbuffer
	sub	eax, [esp+8]			;
	add	eax, [SoundBufBase]		;
	mov	[SoundBufPos], eax		;
	call	TrackSong
	pop	edi

	xor	eax, eax			; Positions (order/row)
	mov	al, [NumRows]
	xor	ecx, ecx
	mov	cl, [RowCountdown]
	sub	eax, ecx
	movzx	ecx, [Order]
	shl	ecx, 16
	or	eax, ecx
	mov	ecx, [NewBufPositions]
	cmp	ecx, 103
	jnb	__NoMoreFreePositions
	mov	[NewBufPositions+8+ecx*8+4], eax
	mov	eax, [SoundBufPos]
	mov	[NewBufPositions+8+ecx*8], eax
	inc	ecx
	mov	[NewBufPositions], ecx
__NoMoreFreePositions:

	mov	eax, [BPMTickDistance]
	mov	[TicksLeft], eax
_NoChannelUpdate:
	pop	ebx
	or	ebx, ebx
	jnz	_MixNextPart

_zbOK:	test	[Flags], flagUseMMX		; Empty MMX state
	jz	_NoEMMS
	emms
_NoEMMS:

	mov	eax, [SoundBuffer]		; Data ready - unlock buffer
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Unlock], eax
						; Finish sync events
	call	_EnterCriticalSection, O SyncSection
	cmp	[SyncBuf3.NumEvents], 0
	je	_NoSyncEvents
	mov	[CurrentSyncBuf], O SyncBuf3
_NoSyncEvents:

	push	esi				; Update positions
	push	edi
	mov	edi, O Buf1Positions
	cmp	[SoundBufBase], 0
	je	__UpdateBuf1P
	mov	edi, O Buf2Positions
__UpdateBuf1P:
	mov	esi, O NewBufPositions
	mov	ecx, [NewBufPositions]
	lea	ecx, [2+ecx+ecx]
__MoveBufP:
	mov	eax, [esi]
	add	esi, 4
	mov	[edi], eax
	add	edi, 4
	dec	ecx
	jnz	__MoveBufP
	pop	edi
	pop	esi
	mov	[LastBeforeBuf], eax
	call	_LeaveCriticalSection, O SyncSection

	;--return--
	ret

_ZeroBuffer:					; Fill remaining part of buffer
	mov	ecx, ebx			; with zeroes
	test	[Flags], flagStereo
	jz	_zbNoStereo
	shl	ecx, 1
_zbNoStereo:
	test	[Flags], flagMix16
	jz	_zbNo16
	shl	ecx, 1
_zbNo16:
	xor	eax, eax
	mov	edx, ecx
	shr	ecx, 2
	and	edx, 03h
	rep	stosd
	mov	ecx, edx
	rep	stosb
	jmp	_zbOK
endp

; Note for both integer and MMX mixing routines:
; optimizing them especially for Cyrix or AMD processors
; gives a little boost, for example on 6x86 133MHz
; with 44kHz stereo mixing optimizing integer routine gave 2% gain.
; On AMD it would not probably be much faster, and the MMX
; routine could not be better optimized.

macro	MixProc procname, interpolate, funct, nchan
local _mix, _mixcontinue, _mixloop, _mixstop, _mixnofill, _aaa
proc	&procname
_mix:
if (interpolate eq 1)
			movsx	ebp, [word esi+esi]
			movsx	eax, [word esi+esi+2]
			shr	ebx, 1
			sub	eax, ebp
			imul	ebx
			add	ebx, ebx
			lea	eax, [ebp+2*edx]
else
			movsx	eax, [word esi+esi]
endif
if (funct eq 1)
			mov	edx, [edi]
else
			mov	edx, [FracStep]
endif
if (nchan eq 2)
			mov	ebp, eax
endif
			imul	eax, [LeftVolume]
if (funct eq 1)
			add	eax, edx
if (nchan eq 2)
			mov	edx, [edi+4]
endif
endif
if (nchan eq 2)
			imul	ebp, [RightVolume]
if (funct eq 1)
			add	ebp, edx
endif
endif
if (funct eq 1)
			mov	edx, [FracStep]
endif
			mov	[edi], eax
if (nchan eq 2)
			mov	[edi+4], ebp
endif
			add	ebx, edx
if (interpolate eq 1)
			adc	esi, 0
else
			adc	esi, [IntStep]
endif
if (nchan eq 2)
			add	edi, 8
else
			add	edi, 4
endif
		; Compare src pointer against offsample pointer
			cmp	esi, [EndPoint]
			jnc	_mixloop

		; Next loop
_mixcontinue:		dec	ecx
			jnz	_mix

		; Finish sample due to mixing buffer fullness
			add	esi, esi
			mov	eax, [ChannelAddr]
			mov	[(CHANNEL eax).SamplePos], esi
			mov	[(CHANNEL eax).FracPos], ebx
			ret

		; Handle case of src pointer overflow
_mixloop:		cmp	[DoLoop], 0
			jz	_mixstop
			sub	esi, [LoopLength]
			jmp	_mixcontinue

		; Stop sample
_mixstop:		mov	ebx, [ChannelAddr]
			and	[(CHANNEL ebx).Flags], not flagSample

if (funct eq 0) ; Fill mixing buffer with zeroes if needed
			dec	ecx
			jz	_mixnofill
if (nchan eq 2)
			shl	ecx, 1
endif
			xor	eax, eax
			rep	stosd
endif
_mixnofill:		ret
endp
endm

MixProc AMixM, 0, 1, 1
MixProc AMixS, 0, 1, 2
MixProc MMixM, 0, 0, 1
MixProc MMixS, 0, 0, 2
MixProc AIntM, 1, 1, 1
MixProc AIntS, 1, 1, 2
MixProc MIntM, 1, 0, 1
MixProc MIntS, 1, 0, 2

; On Pentium MMX 17 cycles without and 22 with interpolation
; Including 5-cycle penalty for misaligned accesses - 22 and 27 cycles
; This is 11 cycles without and 13.5 cycles with interpolation per sample

macro	MMXMixProc procname, interpolate, funct
proc	&procname
if (interpolate eq 1)
			movd	mm0, ebx	; 0 | FracPos
			punpckldq mm0, [FracStep] ; FracStep | FracPos
			psrld	mm0, 17 	; FS>>17 | FP>>17
			pxor	mm5, mm5	; 0
			psubw	mm5, mm0	; 0 | -FS | 0 | -FP
			movq	mm6, mm5	; ditto
			punpcklwd mm5, mm0	; 0 | 0 | FP | -FP
			punpckhwd mm6, mm0	; 0 | 0 | FS | -FS
			psllq	mm6, 32 	; FS | -FS | 0 | 0
			punpckldq mm5, mm5	; FP | -FP | FP | -FP
			paddw	mm5, mm6	; FP+FS | -FP-FS | FP | -FP
			paddw	mm6, mm6	; 2*FS | -2*FS | 0 | 0
			punpckhdq mm6, mm6	; 2*FS | -2*FS | 2*FS | -2*FS
			movq	mm7, [_MMMM]	; 7FFF | 7FFF | 7FFF | 7FFF
			pand	mm5, mm7	; mask off MSBs
			pxor	mm2, mm2	; 0
			movq	mm4, mm5	; b | -b | a | -a
			pcmpeqd mm2, mm4	; compare to 0
			pand	mm2, [_0M0M]	; 0 | 7FFF:0 | 0 | 7FFF:0
			pand	mm4, mm7	; mask off MSBs
			por	mm4, mm2	; max A scale factor if 0
endif
			mov	eax, [FracStep]
			and	eax, 0FFFE0000h ; prevent carries
			mov	ebp, [IntStep]
			shr	ecx, 1
			jnc	@@next
			or	[DoLoop], 4

		; Load first sample
@@next: 		movq	mm0, [esi+esi]		; 1
			add	ebx, eax		; 2
			adc	esi, ebp		; 3
			add	edi, 16 		; -
			cmp	esi, edx		; 4
			jb	@@no_loop_fix_1 	; -

		; Loop sample
			test	[DoLoop], 1
			jz	@@stop1
			sub	esi, [LoopLength]
@@no_loop_fix_1:

if (interpolate eq 1)
		; Interpolation
			punpckldq mm0, [esi+esi]	; 5
			paddw	mm5, mm6		; -
			pmaddwd mm0, mm4		; 6
			pand	mm5, mm7		; -
			pxor	mm2, mm2		; 7
			movq	mm4, mm5		; -
			pcmpeqd mm2, mm5		; 8
			pand	mm2, [_0M0M]		; 9
			psrad	mm0, 15 		; -
			por	mm4, mm2		; 10
			packssdw mm0, mm0		; -
else
			punpcklwd mm0, [esi+esi]	; 5
endif
		; Mix
			punpcklwd mm0, mm0		; 6/11
			add	ebx, eax		; -
			movq	mm2, mm0		; 7/12
			pmullw	mm0, mm3		; -
			adc	esi, ebp		; 8/13
			pmulhw	mm2, mm3		; -
			movq	mm1, mm0		; 10/15
			punpcklwd mm0, mm2		; 11/16
if (funct eq 1)
			paddd	mm0, [edi-16]		; 12/17
endif
			punpckhwd mm1, mm2		; -
if (funct eq 1)
			paddd	mm1, [edi-8]		; 13/18
endif
			movq	[edi-16], mm0		; 14/19
			movq	[edi-8], mm1		; 15/20
			cmp	esi, edx		; 16/21
			jb	@@no_fix_loop_2 	; -

		; Loop sample
			test	[DoLoop], 1
			jz	@@stop2
			sub	esi, [LoopLength]

		; Next iteration
@@no_fix_loop_2:	dec	ecx			; 17/22
			jnz	@@next			; -

		; Even number of interations done
			test	[DoLoop], 4
			jz	@@no_more_samples
			mov	eax, [esi+esi]
			call	@@StoreSample
			add	ebx, [FracStep]
			adc	esi, [IntStep]
			cmp	esi, [EndPoint]
			jb	@@no_more_samples
			test	[DoLoop], 1
			jz	@@stop_and_finish
			sub	esi, [LoopLength]

		; Store used variables
@@no_more_samples:	add	esi, esi
			mov	eax, [ChannelAddr]
			mov	[(CHANNEL eax).SamplePos], esi
			mov	[(CHANNEL eax).FracPos], ebx
			ret

		; Sample ended in the middle of buffer after odd stores
@@stop1:		sub	edi, 16
			movd	eax, mm0
			call	@@StoreSample
			add	ecx, ecx
			test	[DoLoop], 4
			jnz	@@stop_and_finish
			dec	ecx
			jmp	@@stop_and_finish

		; Sample ended in the middle of buffer after even stores
@@stop2:		dec	ecx
			add	ecx, ecx
			test	[DoLoop], 4
			jz	@@stop_and_finish
			inc	ecx

		; Fill the mixing buffer with zeroes and stop sample
@@stop_and_finish:	mov	ebx, [ChannelAddr]
			and	[(CHANNEL ebx).Flags], not flagSample
if (funct eq 0)
			or	ecx, ecx
			jz	@@no_zero_fill
			xor	eax, eax
		    rep stosd
endif
@@no_zero_fill: 	ret

		; Simple sample storing routine
@@StoreSample:
if (interpolate eq 1)
			movsx	ebp, ax
			sar	eax, 16
			sub	eax, ebp
			shr	ebx, 1
			imul	ebx
			add	ebx, ebx
			add	edx, edx
			add	ebp, edx
else
			movsx	ebp, ax
endif
			mov	eax, ebp
			imul	eax, [LeftVolume]
			imul	ebp, [RightVolume]
if (funct eq 1)
			add	[edi], eax
			add	[edi+4], ebp
else
			mov	[edi], eax
			mov	[edi+4], ebp
endif
			add	edi, 16
			ret
endp
endm

ifndef _NOMMX
MMXMixProc AMixMMX, 0, 1
MMXMixProc MMixMMX, 0, 0
MMXMixProc AIntMMX, 1, 1
MMXMixProc MIntMMX, 1, 0
endif

;-----------------------------------------------------------------------------
;	PlaySong(StopEvent) starts playing the module
;-----------------------------------------------------------------------------

_PlaySong@4:
proc	PlaySong aStopEvent
	mov	eax, 1
	test	[Flags], flagNotLoaded + flagPlaying
	jnz	_NothingToStart

	mov	[SpeedCountdown], 1		; Prepare pattern
	mov	[RowCountdown], 1
	mov	[Order], -1
	mov	eax, O ChannelTable
	mov	ecx, [NumChannels]
_ResetNextChannel:
	mov	[(CHANNEL eax).Flags], 0
	mov	[(CHANNEL eax).VibEPos], 0
	mov	[(CHANNEL eax).VibEWave], 0
	mov	[(CHANNEL eax).TremEPos], 0
	mov	[(CHANNEL eax).TremEWave], 0
	add	eax, size CHANNEL
	loop	_ResetNextChannel

	mov	ecx, [aStopEvent]		; Update events
	mov	[StopEvent], ecx
	mov	[CurrentSyncBuf], O SyncBuf2
	xor	ecx, ecx
	mov	[SyncBuf2.NumEvents], ecx
	mov	[SyncBuf1.NumEvents], ecx
	mov	[Buf1Positions], ecx
	mov	[Buf1Positions+4], ecx
	mov	[Buf2Positions], ecx
	mov	[Buf2Positions+4], ecx
	mov	[LastBeforeBuf], ecx

	call	_InitializeCriticalSection, O SyncSection

	push	eax				; Create mixing thread
	mov	eax, esp
	call	_CreateThread, 0, 4000h, O MixingThread, 0, 0, eax
	add	esp, 4
	or	eax, eax
	jz	_pmError1
	mov	[hMixingThread], eax
						; Set thread priority
	test	[Flags], flagHiPriority
	jz	_pmLoPri
	call	_SetThreadPriority, eax, THREAD_PRIORITY_HIGHEST
_pmLoPri:

	mov	eax, [BPMTickDistance]		; Prepare sound buffer
	mov	[TicksLeft], eax
	call	TrackSong
	call	Mix1
	mov	eax, [SoundBuffer]		; Play module
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Play], eax, 0, 0, DSBPLAY_LOOPING
	or	eax, eax
	jnz	_pmError2

	or	[Flags], flagPlaying		; Setup flags
	and	[Flags], not flagLoopModule
	cmp	[aStopEvent], 0
	jnz	_NoLoop
	or	[Flags], flagLoopModule
_NoLoop:
	xor	eax, eax
	ret

_pmError2:					; Error exiting
	call	_TerminateThread, [hMixingThread], 0
	call	_CloseHandle, [hMixingThread]
	mov	eax, 1
_pmError1:
	inc	eax
	inc	eax
	and	[Flags], not flagPlaying
	call	_DeleteCriticalSection, O SyncSection
_NothingToStart:
	ret
endp

;-----------------------------------------------------------------------------
;	StopSong() stops playing the module
;-----------------------------------------------------------------------------

_StopSong@0:
proc	StopSong
	test	[Flags], flagPlaying
	jz	_NothingToStop

	mov	eax, [SoundBuffer]		; Stop playing
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Stop], eax
						; Quit mixing thread
	call	_TerminateThread, [hMixingThread], 0
	call	_CloseHandle, [hMixingThread]
	call	_DeleteCriticalSection, O SyncSection

	and	[Flags], not flagPlaying
	and	[Flags2], not (flag2StopModule+flag2Stopping)
_NothingToStop:
	ret
endp

;-----------------------------------------------------------------------------
;	InitPlayer() function initializes tables
;-----------------------------------------------------------------------------

_InitPlayer@0:
proc	InitPlayer uses ebx edi
	fninit					; Prepare coprocessor for
	push	8363				; frequency table calculation
	fild	[dword esp]
	mov	[dword esp], 1536
	fild	[dword esp]
	fld1
	fld1
	fchs
	mov	ecx, 128*(MAXNOTE+1)
	mov	edi, O FreqTable

_FreqCalc:					; Create frequency table
	fadd	st, st(1)
	fld	st
	fdiv	st, st(3)
	fld	st		; f2x
	frndint 		;
	fsub	st(1), st	;
	fld1			;
	fscale			;
	fstp	st(1)		;
	fxch	st(1)		;
	f2xm1			;
	fld1			;
	faddp	st(1), st	;
	fmulp	st(1), st	;
	fmul	st, st(4)
	fistp	[dword esp]
	mov	eax, [esp]
	shr	eax, 4
	stosd
	loop	_FreqCalc

	fninit					; Prepare to sine calculation
	mov	[dword esp], 255
	fild	[dword esp]
	fldpi
	mov	[dword esp], 128
	fidiv	[dword esp]
	fld	st
	fchs
	mov	ecx, 256

_SinCalc:					; Calculate sine wave
	fadd	st, st(1)
	fld	st
	fsin
	fmul	st, st(3)
	fistp	[dword esp]
	mov	eax, [dword esp]
	stosd
	loop	_SinCalc
	fninit
	add	esp, 4

	mov	cl, 128 			; Square wave
	mov	eax, 255
	rep	stosd
	neg	eax
	mov	cl, 128
	rep	stosd

	mov	ecx, 128			; Ramp down
	xor	eax, eax
_RampDown:
	stosd
	dec	eax
	dec	eax
	loop	_RampDown
	mov	eax, 255
	mov	ecx, 128
_RampDown2:
	stosd
	dec	eax
	dec	eax
	loop	_RampDown2

	mov	ecx, 128			; Ramp up
	xor	eax, eax
_RampUp:
	stosd
	inc	eax
	inc	eax
	loop	_RampUp
	mov	eax, -255
	mov	ecx, 128
_RampUp2:
	stosd
	inc	eax
	inc	eax
	loop	_RampUp2

	ret
endp

;-----------------------------------------------------------------------------
;	LoadSong() function allocates memory and loads the module
;-----------------------------------------------------------------------------

_LoadSong@24:
proc	LoadSong uses ebx esi edi, FName,SName,hwnd,aRate,aFlags,aVolScale
local APatAddrTable[256]: dword
local APatLengthTable[256]: byte
local BufferDesc: DSBUFFERDESC
local WaveFormat: WAVEFORMATEX
local aSampleAddr
local aPingPong: byte, aNumSamples: byte, ErrorCode: byte
	call	UnloadSong		      ; Clear status
	mov	[Flags], flagNotLoaded

	test	[aFlags], lflagUseMMX		; Use MMX if available ?
	jz	_NoMMX

	pushf					; Determine CPUID presence
	pushf
	pop	eax
	mov	ecx, eax
	btc	eax, 21
	push	eax
	popf
	pushf
	pop	eax
	popf
	xor	eax, ecx
	jz	_NoMMX

	xor	eax, eax			; Determine features presence
	cpuid
	or	ax, ax
	jz	_NoMMX

	mov	eax, 1				; Determine MMX presence
	cpuid
	bt	edx, 23
	jnc	_NoMMX

	or	[Flags], flagUseMMX
	or	[aFlags], lflagStereo+lflagMix16
	mov	[aRate], 44100
_NoMMX:
	mov	edi, [FName]			; Check if file is loaded
	test	[aFlags], lflagExternalLoad
	jnz	_AlreadyLoaded
						; Open the file
	call	_CreateFile, [FName], GENERIC_READ, FILE_SHARE_READ,\
			    0, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0

	mov	[ErrorCode], ERROR_FILEOPEN
	cmp	eax, INVALID_HANDLE_VALUE
	jz	_ExitLoadModule
	mov	ebx, eax

	call	_GetFileSize, ebx, 0		; Get size of the file
	mov	[ErrorCode], ERROR_FILESIZE
	cmp	eax, -1
	jz	_ExitLoadModule_CH
	mov	esi, eax

	call	_GlobalAlloc, GMEM_FIXED, esi	; Allocate fixed memory for file
	mov	[ErrorCode], ERROR_MEMORY
	or	eax, eax
	jz	_ExitLoadModule_CH
	mov	edi, eax

	push	eax				; Load the file
	mov	eax, esp
	call	_ReadFile, ebx, edi, esi, eax, 0
	mov	[ErrorCode], ERROR_FILEREAD
	or	eax, eax
	pop	eax
	jz	_ExitLoadModule_CH_GF
	cmp	eax, esi
	jnz	_ExitLoadModule_CH_GF

	call	_CloseHandle, ebx		; Close the file

_AlreadyLoaded:
	mov	ebx, edi			; Check IDText
	mov	esi, O IDText
	mov	ecx, 17
	rep	cmpsb
	mov	[ErrorCode], ERROR_MODULE
	jnz	_ExitLoadModule_GF
	cmp	[(XMHEADER ebx).IDByte], 1Ah	; Check IDByte
	jnz	_ExitLoadModule_GF

	mov	edi, [SName]			; Update user song name
	or	edi, edi
	jz	_lm_NoSongName
	lea	esi, [(XMHEADER ebx).ModuleName]
	mov	ecx, 5
	rep	movsd
	mov	ecx, 20
_lm_ZeroPad:
	mov	[byte edi], 0
	dec	edi
	mov	al, [edi]
	cmp	al, ' '
	jnz	_lm_NoSongName
	loop	_lm_ZeroPad

_lm_NoSongName: 				; Global data
	mov	al, [byte (XMHEADER ebx).SongLength]
	mov	[SongLength], al
	mov	al, [byte (XMHEADER ebx).RestartOrder]
	mov	[RestartOrder], al
	movzx	eax, [(XMHEADER ebx).NumChannels]
	mov	[NumChannels], eax
	mov	al, [byte (XMHEADER ebx).Speed]
	mov	[Speed], al
	mov	eax, [aRate]
	mov	[OutputFreq], eax
	lea	eax, [eax*4+eax]
	xor	edx, edx
	movzx	ecx, [byte (XMHEADER ebx).BPM]
	mov	[BPM], cl
	shl	ecx, 1
	div	ecx
	mov	[BPMTickDistance], eax
	mov	[GlobalVol], 40h

	test	[aFlags], lflagStereo		; Allocate MixingBuffer
	setnz	cl
	add	cl, 2
	shl	eax, cl
	call	_GlobalAlloc, GMEM_FIXED, eax
	mov	[ErrorCode], ERROR_MEMORY
	or	eax, eax
	jz	_ExitLoadModule_GF
	mov	[MixingBuffer], eax

	push	ebx				; Calculate lengths of all pats
	movzx	ecx, [(XMHEADER ebx).NumPatterns]
	lea	esi, [ebx+60]
	add	esi, [esi]
	lea	edi, [APatLengthTable]
	lea	ebx, [APatAddrTable]
	xor	edx, edx
_CalcNextPattern:
	movzx	eax, [(PATTERN esi).NumRows]
	stosb
	mov	[ebx], edx
	add	ebx, 4
	push	edx
	mov	edx, [NumChannels]
	mul	edx
	pop	edx
	shl	eax, 3
	add	edx, eax
	movzx	eax, [(PATTERN esi).PatternSize]
	add	eax, [(PATTERN esi).dwSize]
	add	esi, eax
	loop	_CalcNextPattern
	pop	ebx

	mov	edi, O PatAddrTable		; Update pattern addresses
	mov	esi, O PatLengthTable
	movzx	ecx, [SongLength]
	push	edx
	push	ebx
	mov	dh, [byte (XMHEADER ebx).NumPatterns]
	lea	ebx, [(XMHEADER ebx).OrderTable]
_GetNextPatternAddr:
	movzx	eax, [byte ebx]
	cmp	al, dh
	jc	_PatternNumberOK
	or	eax, -1
	jmp	_WrongPattern
_PatternNumberOK:
	mov	dl, [APatLengthTable+eax]
	mov	[esi], dl
	mov	eax, [APatAddrTable+eax*4]
_WrongPattern:
	stosd
	inc	esi
	inc	ebx
	loop	_GetNextPatternAddr
	pop	ebx
	pop	eax

	call	_GlobalAlloc, GMEM_FIXED, eax	; Allocate memory for patterns
	mov	[ErrorCode], ERROR_MEMORY
	or	eax, eax
	jz	_ExitLoadModule_6
	mov	[PatternMem], eax
	mov	edi, eax
						; Update pattern addresses
	movzx	ecx, [SongLength]
	mov	eax, O PatAddrTable
_UpdatePatAddr:
	add	[eax], edi
	add	eax, 4
	loop	_UpdatePatAddr
						; Allocate memory for instrum_s
	movzx	eax, [(XMHEADER ebx).NumInstruments]
	imul	eax, size INSTRUMENT
	mov	[InstrTableSize], eax
	call	_GlobalAlloc, GMEM_FIXED, eax
	or	eax, eax
	jz	_ExitLoadModule_GF2
	mov	[InstrTable], eax

	lea	esi, [ebx+60]			; Translate patterns
	add	esi, [esi]
	movzx	ecx, [(XMHEADER ebx).NumPatterns]
_XlateNextPattern:
	push	esi
	push	ecx
	movzx	ecx, [(PATTERN esi).NumRows]
	add	esi, [(PATTERN esi).dwSize]

_XlateNextRow:					; Translate rows
	push	ecx
	mov	ecx, [NumChannels]

_XlateNextChannel:				; Translate channels
	mov	[byte edi], 0
	mov	ah, 1Fh 			; Get mask
	test	[byte esi], 80h
	jz	_GetNote
	lodsb
	mov	ah, al

_GetNote:					; Is note ?
	test	ah, pflagNote
	jz	_GetInstr_NoNote

	lodsb					; Get note
	or	al, al
	jz	_GetInstr_NoNote
	dec	al
	cmp	al, 96
	ja	_GetInstr_NoNote
	or	[byte edi], flagResetNote
	mov	[edi+1], al
	cmp	al, 96
	jnz	_GetInstr_IsNote
	test	ah, pflagInstr
	jz	_GetVolume
	inc	esi
	jmp	_GetVolume

_GetInstr_NoNote:				; Get instrument when no note
	test	ah, pflagInstr
	jz	_GetVolume
	lodsb
	or	al, al
	jz	_GetVolume
	or	[byte edi], flagResetVolume
	mov	[byte edi+3], -1
	jmp	_GetVolume

_GetInstr_IsNote:				; Get instrument when is note
	or	[byte edi], flagResetSample
	test	ah, pflagInstr
	jz	_SetJustResetSample
	lodsb
	or	al, al
	jz	_SetJustResetSample
	dec	al
	mov	[edi+2], al
	or	[byte edi], flagResetVolume
	mov	[byte edi+3], -1
	jmp	_GetVolume
_SetJustResetSample:
	mov	[byte edi+2], -1

_GetVolume:					; Get volume
	test	ah, pflagVolume
	jz	_GetEffect
	lodsb
	cmp	al, 10h
	jc	_GetEffect
	cmp	al, 50h
	ja	_GetVolEffect
	sub	al, 10h 			; Real volume
	mov	[edi+3], al
	or	[byte edi], flagResetVolume
	jmp	_GetEffect
_GetVolEffect:
	cmp	al, 60h
	jc	_GetEffect
	sub	al, 60h 			; Volume effect
	movzx	edx, al
	and	al, 0Fh
	shr	dl, 4
	mov	dh, [XlateVolEffect+edx]
	cmp	dh, -1
	jz	_GetEffect
	xchg	dl, cl
	mov	cl, [VolEffectShift+ecx]
	shl	al, cl
	mov	cl, dl
	mov	dl, dh
	mov	dh, al
	mov	[edi+4], dx
	or	[byte edi], flagResetEffect1
	cmp	dl, 3
	jnz	_GetEffect
	and	[byte edi], not flagResetSample

_GetEffect:					; Get effect
	test	ah, pflagEffType
	jz	_GetArpeggio
	lodsb
	xchg	al, ah
	test	al, pflagEffData
	mov	al, 00h
	jz	_XlateEffect
	lodsb
	jmp	_XlateEffect
_GetArpeggio:
	test	ah, pflagEffData
	jz	_ToNextChannel
	lodsb
	mov	ah, 00h
_XlateEffect:
	xchg	al, ah
	or	ax, ax
	jz	_ToNextChannel
	cmp	al, 21h
	ja	_ToNextChannel
	cmp	al, 0Eh
	jnz	_NotEEffect
	mov	al, ah
	shr	al, 4
	and	ah, 0Fh
	add	al, 22h
_NotEEffect:
	movzx	edx, al
	mov	al, [XlateEffect+edx]
	cmp	al, -1
	jz	_ToNextChannel
	test	[byte edi], flagResetEffect1
	jz	_DontOverride
	cmp	al, [edi+4]
	jz	_OverrideEffect1
_DontOverride:
	or	[byte edi], flagResetEffect2
	mov	[edi+6], ax
	cmp	al, 3
	jz	_SkipSampleReset
	cmp	al, 5
	jz	_SkipSampleReset
	cmp	al, 2Fh
	jnz	_ToNextChannel
_SkipSampleReset:
	and	[byte edi], not flagResetSample
	jmp	_ToNextChannel
_OverrideEffect1:
	mov	[edi+5], ah

_ToNextChannel:
	add	edi, 8				; Next channel
	dec	ecx
	jnz	_XlateNextChannel

	pop	ecx				; Next row
	dec	ecx
	jnz	_XlateNextRow

	pop	ecx				; Next pattern
	pop	esi
	movzx	eax, [(PATTERN esi).PatternSize]
	add	eax, [(PATTERN esi).dwSize]
	add	esi, eax
	dec	ecx
	jnz	_XlateNextPattern

	mov	eax, [UserDirectSound]		; Use user's DS is avail
	mov	[DirectSound], eax
	or	eax, eax
	jnz	@@SkipDSCreation

	push	eax				; Create IDirectSound object
	mov	eax, esp
	xor	ecx, ecx
	call	_DirectSoundCreate, ecx, eax, ecx
	mov	[ErrorCode], ERROR_SOUND
	or	eax, eax
	pop	eax
	jnz	_ExitLoadModule_GF_GF
	mov	[DirectSound], eax

@@SkipDSCreation:
	mov	edi, eax
	push	esi
	mov	esi, [edi]
	
	xor	ecx, ecx			; Check if cb enabled
	mov	[PrimaryBuffer], ecx
	mov	ecx, [bCreatePrimary]
	or	ecx, ecx
	jz	_NoPrimaryBuffer1
						; Set cooperative level
	call	[(IDirectSound esi).SetCooperativeLevel], edi, [hwnd], \
							    DSSCL_PRIORITY
	or	eax, eax
	jnz	_ExitLoadModule_GF_GF2

	mov	[BufferDesc.dwSize], 20 	; Create primary sound buffer
	mov	[BufferDesc.dwFlags], DSBCAPS_PRIMARYBUFFER
	mov	[BufferDesc.dwBufferBytes], 0
	mov	[BufferDesc.dwReserved], 0
	mov	[BufferDesc.lpwfxFormat], 0
	mov	[PrimaryBuffer], 0
	lea	eax, [BufferDesc]
	push	ecx
	mov	ecx, esp
	call	[(IDirectSound esi).CreateSoundBuffer], edi, eax, ecx, 0
	or	eax, eax
	pop	eax
	jnz	_NoPrimaryBuffer1
	mov	[PrimaryBuffer], eax

_NoPrimaryBuffer1:				; Prepare format of snd buffers
	mov	[WaveFormat.wFormatTag], WAVE_FORMAT_PCM
	mov	[WaveFormat.cbSize], 0
	mov	eax, [aRate]
	mov	[WaveFormat.nSamplesPerSec], eax
	mov	edx, 1
	test	[aFlags], lflagHiPriority
	jz	_NoHiPriority
	or	[Flags], flagHiPriority
_NoHiPriority:
	test	[aFlags], lflagStereo
	jz	_NoStereo
	or	[Flags], flagStereo
	inc	edx
	shl	eax, 1
_NoStereo:
	mov	[WaveFormat.nChannels], dx
	mov	ecx, 8
	test	[aFlags], lflagMix16
	jz	_NoMix16
	or	[Flags], flagMix16
	shl	dl, 1
	shl	eax, 1
	mov	cl, 16
_NoMix16:
	mov	[WaveFormat.wBitsPerSample], cx
	mov	[WaveFormat.nBlockAlign], dx
	mov	[WaveFormat.nAvgBytesPerSec], eax
	test	[aFlags], lflag1Second
	setnz	cl
	shl	eax, cl
	mov	[BufferDesc.dwBufferBytes], eax
	shr	eax, 1
	mov	[Event2Pos], eax

	mov	eax, [PrimaryBuffer]		; Setup format of primary buffer
	or	eax, eax
	jz	_NoPrimaryBuffer2
	mov	ecx, [eax]
	lea	edx, [WaveFormat]
	call	[(IDirectSoundBuffer ecx).SetFormat], eax, edx
	or	eax, eax
	jz	_NoPrimaryBuffer2
	xor	eax, eax
	xchg	eax, [PrimaryBuffer]
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Release], eax

_NoPrimaryBuffer2:
	lea	eax, [BufferDesc]		; Create mixing buffer
	mov	ecx, DSBCAPS_CTRLPOSITIONNOTIFY + DSBCAPS_GETCURRENTPOSITION2
	test	[aFlags], lflagLocalSound
	jnz	_LocalSound
	or	ecx, DSBCAPS_STICKYFOCUS + DSBCAPS_GLOBALFOCUS
_LocalSound:
	mov	[(DSBUFFERDESC eax).dwFlags], ecx
	lea	edx, [WaveFormat]
	mov	[(DSBUFFERDESC eax).lpwfxFormat], edx
	push	ecx
	mov	ecx, esp
	call	[(IDirectSound esi).CreateSoundBuffer], edi, eax, ecx, 0
	or	eax, eax
	pop	eax
	pop	esi
	jnz	_ExitLoadModule_1
	mov	[SoundBuffer], eax

	xor	eax, eax			; Create 1st notification event
	call	_CreateEvent, eax, eax, eax, eax
	mov	[ErrorCode], ERROR_EVENT
	or	eax, eax
	jz	_ExitLoadModule_2
	mov	[Event1], eax
	mov	[dword Events], eax

	xor	eax, eax			; Create 2nd notification event
	call	_CreateEvent, eax, eax, eax, eax
	or	eax, eax
	jz	_ExitLoadModule_3
	mov	[Event2], eax
	mov	[dword Events+4], eax

	push	edx				; Create notification object
	mov	edx, esp
	mov	eax, [SoundBuffer]
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).QueryInterface], eax,\
						O IID_IDirectSoundNotify, edx
	mov	[ErrorCode], ERROR_NOTIFY
	or	eax, eax
	pop	eax
	jnz	_ExitLoadModule_4
	mov	[BufferNotify], eax

	mov	[Event1Pos], 0			; Set up notification
	mov	ecx, [eax]
	call	[(IDirectSoundNotify ecx).SetNotificationPositions],\
							eax, 2, O Notifies

	mov	edx, [InstrTable]		; Clear instruments
	movzx	ecx, [(XMHEADER ebx).NumInstruments]
	mov	eax, -1
_ClearNextInstrument:
	push	ecx
	mov	[(INSTRUMENT edx).Flags], 0
	lea	edi, [(INSTRUMENT edx).Samples]
	mov	ecx, 24
	rep	stosd
	pop	ecx
	add	edx, size INSTRUMENT
	loop	_ClearNextInstrument

	push	esi				; Calculate size of memory
	push	ebx				; for samples
	xor	edi, edi
	movzx	ecx, [(XMHEADER ebx).NumInstruments]
	xor	ebx, ebx
_CalcNextInstrument:
	push	ecx
	movzx	ecx, [(INSTRUMENTHEADER esi).NumSamples]
	mov	edx, [(INSTRUMENTHEADER esi).DwSampleHeaderSize]
	add	esi, [(INSTRUMENTHEADER esi).dwSize]
	or	ecx, ecx
	jz	_SkipCalcInstr
	xor	eax, eax
_CalcNextSample:
	push	ecx
	cmp	[(SAMPLEHEADER esi).SampleLength], 0
	jz	_SkipCalcSample
	inc	ebx
	mov	ecx, [(SAMPLEHEADER esi).SampleLength]
	add	eax, ecx
	test	[(SAMPLEHEADER esi).Type], 03h
	jz	_AddFinalLength 		; Sample with loop
	mov	ecx, [(SAMPLEHEADER esi).SampleLoopBegin]
	add	ecx, [(SAMPLEHEADER esi).SampleLoopLength]
	test	[(SAMPLEHEADER esi).Type], 02h
	jz	_AddFinalLength
	add	ecx, [(SAMPLEHEADER esi).SampleLoopLength]
_AddFinalLength:
	test	[(SAMPLEHEADER esi).Type], 10h	; Convert all 8bit to 16bit
	jnz	_NoConv
	shl	ecx, 1
_NoConv:
	lea	edi, [edi+ecx+2]		; Add final sample length
_SkipCalcSample:
	add	esi, edx
	pop	ecx
	loop	_CalcNextSample
	add	esi, eax
_SkipCalcInstr:
	pop	ecx
	loop	_CalcNextInstrument
	mov	eax, ebx
	pop	ebx
	pop	esi

	mov	[aNumSamples], al		; Allocate memory for samples
	imul	eax, size SAMPLE
	mov	[SampleTableSize], eax
	lea	eax, [eax+edi+10]	; to prevent GPF/NP while mixing
	call	_GlobalAlloc, GMEM_FIXED, eax
	mov	[ErrorCode], ERROR_MEMORY
	or	eax, eax
	jz	_ExitLoadModule_5
	mov	[SampleBuffer], eax
	add	eax, 2
	mov	[aSampleAddr], eax
	lea	eax, [eax+edi+8]
	mov	[SampleTable], eax

	mov	edi, eax			; Clear samples
	movzx	ecx, [aNumSamples]
	imul	ecx, (size SAMPLE) shr 2
	xor	eax, eax
	rep	stosd
						; Translate instruments
	movzx	ecx, [(XMHEADER ebx).NumInstruments]
	mov	edi, [InstrTable]
	mov	[aNumSamples], 0
_XlateNextInstrument:
	push	ecx
	cmp	[(INSTRUMENTHEADER esi).NumSamples], 0
	jnz	_GetVolumeEnv
	add	esi, [(INSTRUMENTHEADER esi).dwSize]
	jmp	_SkipInstrument
_GetVolumeEnv:					; Get volume envelope
	test	[(INSTRUMENTHEADER esi).VolType], 1
	jz	_NoVolEnv
						; Get vol env border
	or	[(INSTRUMENT edi).Flags], flagVolEnvelope
	movzx	eax, [(INSTRUMENTHEADER esi).VESize]
	dec	eax
	mov	ax, [((INSTRUMENTHEADER esi).VolumeEnvelope)+eax*4]
	inc	ax
	mov	[(INSTRUMENT edi).VolEnvBorder], ax
						; Get vol env sustain
	test	[(INSTRUMENTHEADER esi).VolType], 02h
	jnz	_VolSustain
	mov	[(INSTRUMENT edi).VolSustain], -1
	jmp	_GetVolLoop
_VolSustain:
	movzx	eax, [(INSTRUMENTHEADER esi).VolSustain]
	mov	ax, [((INSTRUMENTHEADER esi).VolumeEnvelope)+eax*4]
	mov	[(INSTRUMENT edi).VolSustain], ax

_GetVolLoop:					; Get vol env loop
	test	[(INSTRUMENTHEADER esi).VolType], 04h
	jnz	_VolLoop
	mov	[(INSTRUMENT edi).VolEndP1], -1
	jmp	_GetVolEnvPoints
_VolLoop:
	movzx	eax, [(INSTRUMENTHEADER esi).VolLoopBegin]
	mov	ax, [((INSTRUMENTHEADER esi).VolumeEnvelope)+eax*4]
	mov	[(INSTRUMENT edi).VolBegin], ax
	movzx	eax, [(INSTRUMENTHEADER esi).VolLoopEnd]
	mov	ax, [((INSTRUMENTHEADER esi).VolumeEnvelope)+eax*4]
	inc	ax
	mov	[(INSTRUMENT edi).VolEndP1], ax

_GetVolEnvPoints:				; Get vol env chart
	movzx	ecx, [(INSTRUMENTHEADER esi).VESize]
	dec	ecx
	jnz	_NoEmptyVolEnv
	mov	al, [byte ((INSTRUMENTHEADER esi).VolumeEnvelope)+2]
	mov	[(INSTRUMENT edi).VolEnv], al
	jmp	_NoVolEnv
_NoEmptyVolEnv:
	push	ebx
	push	esi
	push	edi
	lea	esi, [(INSTRUMENTHEADER esi).VolumeEnvelope]
	lea	edi, [(INSTRUMENT edi).VolEnv]
	mov	al, [esi+2]
	stosb
_NextVolEnvPoint:
	push	ecx				; Prepare to line
	add	esi, 4
	push	esi
	mov	dx, [esi]
	sub	dx, [esi-4]
	mov	cx, dx
	mov	bx, [esi+2]
	sub	bx, [esi-2]
	mov	si, 0
	jc	_VolEnvDown
_VEUpNext:					; Line up
	add	si, bx
_VEUNCheck:
	cmp	si, dx
	jl	_VEUNSet
	sub	si, dx
	inc	al
	jmp	_VEUNCheck
_VEUNSet:
	stosb
	loop	_VEUpNext
	jmp	_DoneVE
_VolEnvDown:					; Line down
	neg	bx
_VEDnNext:
	add	si, bx
_VEDNCheck:
	cmp	si, dx
	jl	_VEDNSet
	sub	si, dx
	dec	al
	jmp	_VEDNCheck
_VEDNSet:
	stosb
	loop	_VEDnNext
_DoneVE:					; Vol env chart ready
	pop	esi
	pop	ecx
	loop	_NextVolEnvPoint
	pop	edi
	pop	esi
	pop	ebx
_NoVolEnv:					; Get panning envelope
	test	[(INSTRUMENTHEADER esi).PanType], 1
	jz	_NoPanEnv
						; Get pan env border
	or	[(INSTRUMENT edi).Flags], flagPanEnvelope
	movzx	eax, [(INSTRUMENTHEADER esi).PESize]
	dec	eax
	mov	ax, [((INSTRUMENTHEADER esi).PanningEnvelope)+eax*4]
	inc	ax
	mov	[(INSTRUMENT edi).PanEnvBorder], ax
						; Get pan env sustain
	test	[(INSTRUMENTHEADER esi).PanType], 02h
	jnz	_PanSustain
	mov	[(INSTRUMENT edi).PanSustain], -1
	jmp	_GetPanLoop
_PanSustain:
	movzx	eax, [(INSTRUMENTHEADER esi).PanSustain]
	mov	ax, [((INSTRUMENTHEADER esi).PanningEnvelope)+eax*4]
	mov	[(INSTRUMENT edi).PanSustain], ax

_GetPanLoop:					; Get pan env loop
	test	[(INSTRUMENTHEADER esi).PanType], 04h
	jnz	_PanLoop
	mov	[(INSTRUMENT edi).PanEndP1], -1
	jmp	_GetPanEnvPoints
_PanLoop:
	movzx	eax, [(INSTRUMENTHEADER esi).PanLoopBegin]
	mov	ax, [((INSTRUMENTHEADER esi).PanningEnvelope)+eax*4]
	mov	[(INSTRUMENT edi).PanBegin], ax
	movzx	eax, [(INSTRUMENTHEADER esi).PanLoopEnd]
	mov	ax, [((INSTRUMENTHEADER esi).PanningEnvelope)+eax*4]
	inc	ax
	mov	[(INSTRUMENT edi).PanEndP1], ax

_GetPanEnvPoints:				; Get pan env chart
	movzx	ecx, [(INSTRUMENTHEADER esi).PESize]
	dec	ecx
	jnz	_NoEmptyPanEnv
	mov	al, [byte ((INSTRUMENTHEADER esi).PanningEnvelope)+2]
	mov	[(INSTRUMENT edi).PanEnv], al
	jmp	_NoPanEnv
_NoEmptyPanEnv:
	push	ebx
	push	esi
	push	edi
	lea	esi, [(INSTRUMENTHEADER esi).PanningEnvelope]
	lea	edi, [(INSTRUMENT edi).PanEnv]
	mov	al, [esi+2]
	stosb
_NextPanEnvPoint:
	push	ecx				; Prepare to line
	add	esi, 4
	push	esi
	mov	dx, [esi]
	sub	dx, [esi-4]
	mov	cx, dx
	mov	bx, [esi+2]
	sub	bx, [esi-2]
	mov	si, 0
	jc	_PanEnvDown
_PEUpNext:					; Line up
	add	si, bx
_PEUNCheck:
	cmp	si, dx
	jl	_PEUNSet
	sub	si, dx
	inc	al
	jmp	_PEUNCheck
_PEUNSet:
	stosb
	loop	_PEUpNext
	jmp	_DonePE
_PanEnvDown:					; Line down
	neg	bx
_PEDnNext:
	add	si, bx
_PEDNCheck:
	cmp	si, dx
	jl	_PEDNSet
	sub	si, dx
	dec	al
	jmp	_PEDNCheck
_PEDNSet:
	stosb
	loop	_PEDnNext
_DonePE:					; Pan env chart ready
	pop	esi
	pop	ecx
	loop	_NextPanEnvPoint
	pop	edi
	pop	esi
	pop	ebx
_NoPanEnv:					; Get vibrato and volume fadeout
	mov	al, [(INSTRUMENTHEADER esi).VibDepth]
	mov	[(INSTRUMENT edi).VibDepth], al
	mov	al, [(INSTRUMENTHEADER esi).VibRate]
	mov	[(INSTRUMENT edi).VibSpeed], al
	mov	al, [(INSTRUMENTHEADER esi).VibSweep]
	mov	[(INSTRUMENT edi).VibSweep], al
	movzx	eax, [(INSTRUMENTHEADER esi).VibType]
	shl	eax, 10
	add	eax, O VibTable
	mov	[(INSTRUMENT edi).VibWaveAddr], eax
	movzx	eax, [(INSTRUMENTHEADER esi).Fadeout]
	mov	[(INSTRUMENT edi).Fadeout], eax

	push	edi				; Clear sample dispatch table
	lea	edi, [(INSTRUMENT edi).Samples]
	xor	ecx, ecx
	xor	eax, eax
	mov	cl, 96/4
    rep stosd

	push	ebx				; Prepare to load samples
	push	esi
	movzx	ecx, [(INSTRUMENTHEADER esi).NumSamples]
	mov	eax, ecx
	mul	[(INSTRUMENTHEADER esi).DwSampleHeaderSize]
	add	esi, [(INSTRUMENTHEADER esi).dwSize]
	mov	edi, esi
	add	esi, eax
	movzx	ebx, [aNumSamples]
	imul	ebx, size SAMPLE
	add	ebx, [SampleTable]
_XlateNextSample:
	push	ecx				; Is sample for that entry ?
	cmp	[(SAMPLEHEADER edi).SampleLength], 0
	jz	_TryNextSample
						; Get basic sample parameters
	mov	al, [(SAMPLEHEADER edi).Volume]
	mov	[(SAMPLE ebx).Volume], al
	mov	al, [(SAMPLEHEADER edi).Panning]
	mov	[(SAMPLE ebx).Panning], al
	movsx	eax, [(SAMPLEHEADER edi).RelativeNote]
	shl	eax, 7
	movsx	ecx, [(SAMPLEHEADER edi).Finetune]
	add	eax, ecx
	mov	[(SAMPLE ebx).Finetune], eax
	mov	[(SAMPLE ebx).Flags], 0

	test	[(SAMPLEHEADER edi).Type], 10h	; Update sample header for 8bit
	jnz	_No8bitUpdate
	shl	[(SAMPLEHEADER edi).SampleLength], 1
	shl	[(SAMPLEHEADER edi).SampleLoopBegin], 1
	shl	[(SAMPLEHEADER edi).SampleLoopLength], 1
_No8bitUpdate:

	mov	[aPingPong], 0			; Prepare loop data
	mov	eax, [(SAMPLEHEADER edi).SampleLength]
	test	[(SAMPLEHEADER edi).Type], 03h
	jz	_NoSampleLoop
	mov	eax, [(SAMPLEHEADER edi).SampleLoopBegin]
	mov	ecx, [(SAMPLEHEADER edi).SampleLoopLength]
	or	[(SAMPLE ebx).Flags], flagLoop
	test	[(SAMPLEHEADER edi).Type], 02h
	jz	_NotPingPongLoop
	mov	[aPingPong], 1
	lea	ecx, [ecx+ecx-4]    ; 1 2..n-1 n n-1..2 (and back to 1)
_NotPingPongLoop:
	add	eax, ecx
	mov	[(SAMPLE ebx).LoopLength], ecx
_NoSampleLoop:
	mov	edx, [aSampleAddr]
	mov	[(SAMPLE ebx).SampleBuf], edx
	add	eax, edx
	mov	[(SAMPLE ebx).LoopEnd], eax
	add	eax, 2
	mov	[aSampleAddr], eax
						; Copy sample bytes
	mov	ecx, [(SAMPLEHEADER edi).SampleLength]
	test	[(SAMPLEHEADER edi).Type], 10h
	jnz	_NoRestore8bitLength
	shr	[(SAMPLEHEADER edi).SampleLength], 1
_NoRestore8bitLength:
	test	[(SAMPLEHEADER edi).Type], 03h
	jz	_DontCalcLoop
	mov	ecx, [(SAMPLEHEADER edi).SampleLoopBegin]
	add	ecx, [(SAMPLEHEADER edi).SampleLoopLength]
_DontCalcLoop:
	shr	ecx, 1
	test	[(SAMPLEHEADER edi).Type], 10h
	push	esi
	push	edi
	mov	edi, edx
	mov	edx, 0
	mov	eax, 0
	jnz	_16bitLoop
_8bitLoop:
	mov	ah, [esi]
	inc	esi
	add	eax, edx
	mov	edx, eax
	stosw
	loop	_8bitLoop
	jmp	_EndMove
_16bitLoop:
	lodsw
	add	eax, edx
	mov	edx, eax
	stosw
	loop	_16bitLoop
_EndMove:

	cmp	[aPingPong], 0			; Finish pingpong loop
	jz	_DontFinishPingPong
	mov	ecx, [(SAMPLE ebx).LoopLength]
	shr	ecx, 2
	sub	ecx, 2
	lea	esi, [edi-4]
_MovePingPong:
	movsw
	sub	esi, 4
	loop	_MovePingPong
_DontFinishPingPong:

	test	[(SAMPLE ebx).Flags], flagLoop	; Update last sample word
	jz	_LastSample			; for interpolation
	mov	eax, [(SAMPLE ebx).LoopEnd]
	sub	eax, [(SAMPLE ebx).LoopLength]
	mov	ax, [eax]
_LastSample:
	stosw
	pop	edi
	pop	esi

	push	ebx				; Update dispatch table
	mov	ebx, [esp+8]
	mov	ah, [byte (INSTRUMENTHEADER ebx).NumSamples]
	sub	ah, [esp+4]
	mov	al, [aNumSamples]
	mov	ecx, 96
	mov	edx, [esp+16]
	lea	edx, [(INSTRUMENT edx).Samples]
	lea	ebx, [(INSTRUMENTHEADER ebx).SampleDispatch]
_UpdateDispatch:
	cmp	[ebx], ah
	jnz	_NextDispatchEntry
	mov	[edx], al
_NextDispatchEntry:
	inc	edx
	inc	ebx
	loop	_UpdateDispatch
	pop	ebx

	add	ebx, size SAMPLE
	inc	[aNumSamples]

_TryNextSample: 				; Next sample
	pop	ecx
	mov	eax, [esp]
	add	esi, [(SAMPLEHEADER edi).SampleLength]
	add	edi, [(INSTRUMENTHEADER eax).DwSampleHeaderSize]
	dec	ecx
	jnz	_XlateNextSample
	pop	ebx	;<- recent instrheader
	pop	ebx
	pop	edi

_SkipInstrument:				; Next instrument
	pop	ecx
	add	edi, size INSTRUMENT
	dec	ecx
	jnz	_XlateNextInstrument

	mov	eax, [aVolScale]		; Get volume scaling factor
	mov	[VolScale], eax
						; Successful return
	and	[Flags], flagStereo+flagMix16+flagUseMMX+flagHiPriority
	test	[aFlags], lflagExternalLoad
	jnz	_NoModMemFree
	call	_GlobalFree, ebx
_NoModMemFree:
	xor	eax, eax
	ret

_ExitLoadModule_CH_GF:				; Error exit
	call	_GlobalFree, edi
_ExitLoadModule_CH:
	call	_CloseHandle, ebx
	xor	eax, eax
	ret

_ExitLoadModule_GF_GF2:
	pop	esi
	jmp	_ExitLoadModule_0
_ExitLoadModule_5:
	mov	eax, [BufferNotify]
	mov	ecx, eax
	call	[(IDirectSoundNotify ecx).Release], eax
_ExitLoadModule_4:
	call	_CloseHandle, [Event2]
_ExitLoadModule_3:
	call	_CloseHandle, [Event1]
_ExitLoadModule_2:
	mov	eax, [SoundBuffer]
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Release], eax
_ExitLoadModule_1:
	mov	eax, [PrimaryBuffer]
	or	eax, eax
	jz	_ExitLoadModule_0
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Release], eax
_ExitLoadModule_0:
	mov	eax, [DirectSound]
	mov	ecx, [eax]
	call	[(IDirectSound ecx).Release], eax
_ExitLoadModule_GF_GF:
	call	_GlobalFree, [InstrTable]
_ExitLoadModule_GF2:
	call	_GlobalFree, [PatternMem]
_ExitLoadModule_6:
	call	_GlobalFree, [MixingBuffer]
_ExitLoadModule_GF:
	test	[aFlags], lflagExternalLoad
	jnz	_ExitLoadModule
	call	_GlobalFree, ebx
_ExitLoadModule:
	movzx	eax, [ErrorCode]
	ret
endp

;-----------------------------------------------------------------------------
;	UnloadSong() just releases allocated memory and DirectX objects
;-----------------------------------------------------------------------------

_UnloadSong@0:
proc	UnloadSong uses ebx esi edi
	test	[Flags], flagNotLoaded
	jnz	_NothingToUnload
	call	StopSong

	mov	eax, [BufferNotify]		; Release DirectSndNotify object
	mov	ecx, [eax]
	call	[(IDirectSoundNotify ecx).Release], eax

	mov	eax, [SoundBuffer]		; Release secondary SoundBuffer
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Release], eax

	mov	eax, [PrimaryBuffer]		; Release primary SoundBuffer
	or	eax, eax
	jz	@@NoPrimaryToRelease
	mov	ecx, [eax]
	call	[(IDirectSoundBuffer ecx).Release], eax

@@NoPrimaryToRelease:
	mov	eax, [UserDirectSound]
	or	eax, eax
	jnz	@@DontReleaseDS
	mov	eax, [DirectSound]		; Release DirectSound object
	mov	ecx, [eax]
	call	[(IDirectSound ecx).Release], eax
@@DontReleaseDS:

	call	_CloseHandle, [Event1]		; Release events
	call	_CloseHandle, [Event2]

	call	_GlobalFree, [SampleBuffer]	; Release memory
	call	_GlobalFree, [InstrTable]
	call	_GlobalFree, [PatternMem]
	call	_GlobalFree, [MixingBuffer]

	or	[Flags], flagNotLoaded
_NothingToUnload:
	ret
endp

;-----------------------------------------------------------------------------
;	GetSongPosition() returns module position (hi=order, lo=row)
;-----------------------------------------------------------------------------

_GetSongPosition@0:
proc	GetSongPosition
		; Check if playing
			test	[Flags], flagPlaying
			jz	@@ReturnFALSE

		; Enter critical section
			call	_EnterCriticalSection, O SyncSection

		; Get position within sound buffer
			mov	eax, [SoundBuffer]
			mov	ecx, [eax]
			sub	esp, 4
			mov	edx, esp
			call	[(IDirectSoundBuffer ecx).GetCurrentPosition],\
								eax, edx, 0
			pop	ecx
			cmp	eax, DS_OK
			jne	@@LeaveAndReturnFALSE

		; Choose sound buffer
			mov	eax, O Buf1Positions
			cmp	ecx, [Event2Pos]
			jb	@@InBuf1
			mov	eax, O Buf2Positions
@@InBuf1:
		; Get number of positions
			mov	edx, [eax]
			add	eax, 8

		; Determine order/row
			or	edx, edx
			jz	@@PositionFound
@@Seek: 		cmp	ecx, [eax]
			jb	@@PositionFound
			add	eax, 8
			dec	edx
			jnz	@@Seek

		; Return order/row
@@PositionFound:	mov	ecx, [eax-4]
			push	ecx
			call	_LeaveCriticalSection, O SyncSection
			pop	eax
			ret

		; Return FALSE
@@LeaveAndReturnFALSE:	call	_LeaveCriticalSection, O SyncSection
@@ReturnFALSE:		or	eax, -1
			ret
endp

;-----------------------------------------------------------------------------
;	GetSyncEventState() - returns state of the last event or -1
;-----------------------------------------------------------------------------

_GetSyncEventState@0:
proc	GetSyncEventState uses esi edi
		; Check if playing
			test	[Flags], flagPlaying
			jz	@@ReturnFALSE

		; Lock event buffers
			call	_EnterCriticalSection, O SyncSection

		; Get position within sound buffer
			mov	eax, [SoundBuffer]
			mov	ecx, [eax]
			sub	esp, 4
			mov	edx, esp
			call	[(IDirectSoundBuffer ecx).GetCurrentPosition],\
								eax, edx, 0
			pop	ecx
			cmp	eax, DS_OK
			jne	@@LeaveAndReturnFALSE

		; Get buffer ID
			xor	edx, edx
			cmp	ecx, [Event2Pos]
			setnb	dl
			inc	edx

		; Skip buffer which isn't played yet
			mov	edi, [CurrentSyncBuf]
			cmp	[(SYNCBUF edi).BufferID], edx
			je	@@NoSkip
			sub	edi, size SYNCBUF
@@NoSkip:
		; Find any event in the current buffer
			mov	edx, [(SYNCBUF edi).NumEvents]
			or	edx, edx
			je	@@CheckOldBuffer
			xor	esi, esi
@@FindEvent:		cmp	[dword ((SYNCBUF edi).Events) + esi*8], ecx
			jnb	@@EventFound
			inc	esi
			cmp	esi, edx
			jb	@@FindEvent

		; Return the event from the current buffer
@@EventFound:		dec	esi
			js	@@CheckOldBuffer
			or	eax, -1
			xchg	eax, [dword ((SYNCBUF edi).Events) + esi*8+4]
			cmp	eax, -1
			je	@@LeaveAndReturnFALSE
@@LeaveAndReturnEvent:	push	eax
			call	_LeaveCriticalSection, O SyncSection
			pop	eax
			ret

		; Check old buffer
@@CheckOldBuffer:	cmp	edi, O SyncBuf1
			jbe	@@LeaveAndReturnFALSE
			sub	edi, size SYNCBUF
			mov	edx, [(SYNCBUF edi).NumEvents]
			dec	edx
			js	@@LeaveAndReturnFALSE
			or	eax, -1
			xchg	eax, [dword ((SYNCBUF edi).Events) + edx*8+4]
			cmp	eax, -1
			jne	@@LeaveAndReturnEvent

		; Return FALSE
@@LeaveAndReturnFALSE:	call	_LeaveCriticalSection, O SyncSection
@@ReturnFALSE:		or	eax, -1
			ret
endp

;-----------------------------------------------------------------------------
;	SetDirectSound() - overrides standard DirectSound object
;-----------------------------------------------------------------------------

_SetDirectSound@8:
proc	SetDirectSound, NewDirectSound, CreatePrimary
			mov	eax, [NewDirectSound]
			mov	[UserDirectSound], eax
			mov	eax, [CreatePrimary]
			mov	[bCreatePrimary], eax
			ret
endp

end