
;	This is an XM player module header file. The player is designed
;	for Win32. It is for assembly use, but can be used with C.
;	An example on how to use this player:
;		call	InitPlayer
;		call	LoadSong, offset ModuleFileName, offset SongName,\
;				  [hwnd], 44100, 11, 4
;		or	eax, eax
;		jz	lb_02
;		call	PlaySong, 0
;		or	eax, eax
;		jz	lb_01
;		; blah, blah - maybe message loop
;		call	StopSong
;	lb_01:	call	UnloadSong
;	lb_02:

; ********* Load flags

xmfStereo		= 1	; mix to stereo instead of mono
xmf16bit		= 2	; mix to 16bit instead of 8bit
xmfLongBuffer		= 4	; use 1 second mixing buffer instead of 0.5s
xmfUseMMX		= 8	; use MMX if available
xmfHiPriority		= 16	; create mixing thread with high priority
xmfExternalLoad 	= 32	; module in memory, pointer in lpszModuleName
xmfLocalSound		= 64	; sound is audible only in current application

; ********* Load return error codes

XM_NOERROR		= 0
XM_ERROR_FILEOPEN	= 1	; Could not open file
XM_ERROR_FILESIZE	= 2	; Could not get file size
XM_ERROR_FILEREAD	= 3	; Could not read file
XM_ERROR_MEMORY 	= 4	; Could not allocate memory
XM_ERROR_MODULE 	= 5	; Error in module file (not XM?)
XM_ERROR_SOUND		= 6	; DirectSound error (eg. no sound card)
XM_ERROR_EVENT		= 7	; Could not create event
XM_ERROR_NOTIFY 	= 8	; DirectSoundNotify error

; ********* Play return error codes

XM_ERROR_PLAYCALL	= 1	; Module not loaded or already playing
XM_ERROR_THREAD 	= 2	; Could not create mixing thread
XM_ERROR_PLAY		= 3	; Could not play mixing dsbuffer

; ********* Functions

; Init player data (called only once before call to any other function)
procdesc InitPlayer

; Load module from file
procdesc LoadSong :dword, :dword, :dword, :dword, :dword, :dword
; parameters:	byte * lpszModuleFileName	module filename (or pointer)
;		byte * lpszSongName		returns song name if non-null
;		dword hwnd			application window
;		dword dwFreq			mixing frequency
;		dword xmfFlags			load flags (see above)
;		dword dwVol			Volume multiplier (best is 7)
; returns load error code or XM_NOERROR == 0
; If you want to load a module yourself to memory, pass a pointer
; to it in lpszModuleFileName and set an apropriate flag in xmfFlags,
; i.e. xmfExternalLoad.

; Play module
procdesc PlaySong :dword
; parameters:	dword hEvent
;     If hEvent is 0 the module will restart at the end, otherwise it will
;	stop and set the event state to signaled.
; returns play error code or XM_NOERROR == 0

; Releasing
procdesc StopSong
procdesc UnloadSong
; Note: you need not to call StopSong() before UnloadSong(), because
; the second one calls the first one if a module is loaded and playing

; Info
procdesc GetSongPosition
; returns module position. High word is order number and low word is
; row number. Returns -1 if unable to retrieve position (e.g. the
; module wasn't started).
procdesc GetSyncEventState
; Returns data of the last synchronization command (Wxx), or -1 if no
; such command have been encountered since the last call to this
; function.

; Overriding
procdesc SetDirectSound :dword, :dword
; Supplies user's DirectSound object in first argument; can be 0 to
; allow standard behavior. The second argument (bool) specifies whether
; to create a primary sound buffer or not.
