/*
	This is an XM player module header file. The player is designed
	for Win32. It is for assembly use, but can be used with C.
	An example on how to use this player:
		InitPlayer();
		if (LoadSong(&szModuleFileName,0,hwnd,44100,11,4) == 0)
		 {
		   if (PlaySong(0) == 0)
		    {
		      // blah, blah - maybe message loop
		      StopSong();
		    };
		   UnloadSong();
		 };
*/

#ifndef __XM_PLAYER
#define __XM_PLAYER

// ********* Load flags

#define xmfStereo	1	// mix to stereo instead of mono
#define xmf16bit	2	// mix to 16bit instead of 8bit
#define xmfLongBuffer	4	// use 1 second mixing buffer instead of 0.5s
#define xmfUseMMX	8	// use MMX if available
#define xmfHiPriority	16	// create mixing thread with high priority
#define xmfExternalLoad 32	// module in memory, pointer in lpszModuleName
#define xmfLocalSound	64	// sound is audible only in current application

// ********* Load return error codes

#define XM_NOERROR		0
#define XM_ERROR_FILEOPEN	1	// Could not open file
#define XM_ERROR_FILESIZE	2	// Could not get file size
#define XM_ERROR_FILEREAD	3	// Could not read file
#define XM_ERROR_MEMORY 	4	// Could not allocate memory
#define XM_ERROR_MODULE 	5	// Error in module file (not XM?)
#define XM_ERROR_SOUND		6	// DirectSound error (eg. no sound card)
#define XM_ERROR_EVENT		7	// Could not create event
#define XM_ERROR_NOTIFY 	8	// DirectSoundNotify error

// ********* Play return error codes

#define XM_ERROR_PLAYCALL	1	// Module not loaded or already playing
#define XM_ERROR_THREAD 	2	// Could not create mixing thread
#define XM_ERROR_PLAY		3	// Could not play mixing dsbuffer

// ********* Functions

#ifdef __cplusplus
extern "C" {
#endif

// Init player data (called only once before any call to other functions)
void __stdcall InitPlayer ( void );

// Load module from file
int __stdcall LoadSong ( char * lpszModuleFileName, // can be a memory pointer
		       char * lpszSongName, // this can be NULL
		       HWND hwnd,      // application window
		       long dwFreq,    // mixing frequency
		       long xmfFlags,  // load flags (see above)
		       long dwVol);    // Volume multiplier (best is 7)
// If you want to load a module yourself to memory, pass a pointer
// to it in lpszModuleFileName (e.g. (char *)MyPtr), and set an
// apropriate flag in xmfFlags, i.e. xmfExternalLoad.

// Play module
int __stdcall PlaySong ( HANDLE hFinishEvent );
// If hEvent is 0 the module will restart at the end, otherwise it will
//   stop and set the event state to signaled.
// returns play error code or XM_NOERROR == 0

// Releasing
int __stdcall StopSong ( void );
int __stdcall UnloadSong ( void );
// Note: you need not to call StopSong() before UnloadSong(), because
// the second one calls the first one if a module is loaded and playing.

// Info
int __stdcall GetSongPosition ( void );
// returns module position. High word is order number and low word is
// row number. Returns -1 if unable to retrieve position (e.g. the
// module wasn't started).
int __stdcall GetSyncEventState ( void );
// Returns data of the last synchronization command (Wxx), or -1 if no
// such command have been encountered since the last call to this
// function.

// Overriding
void __stdcall SetDirectSound ( void* lpDirectSound, BOOL bCreatePrimary );
// Supplies user's DirectSound object in first argument; can be 0 to
// allow standard behavior. The second argument (bool) specifies whether
// to create a primary sound buffer or not.

#ifdef __cplusplus
};//extern "C"
#endif

#endif
